# -*- coding: utf-8 -*-

# Copyright © 2019, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from math import pi
import numpy as np
import pandas as pd
from scipy.spatial import ConvexHull
from collections import OrderedDict


setup = {'name': 'spherical.volume',
        'provides': 'volume'}

def ritter(points):
    if len(points) == 1:
        return points, 0.

    eps = 1e-8

    i = 0
    x = points[[i]]
    dx = points - x
    j = np.argmax(np.sum(dx * dx, axis=1))
    y = points[[j]]
    dy = points - y
    k = np.argmax(np.sum(dy * dy, axis=1))
    z = points[[k]]
    center = .5 * (y + z)
    yz = z - y
    radius = .5 * np.sqrt(np.sum(yz * yz)) + eps

    covered = set((j,k))
    while True:
        dr = points - center
        d2 = np.sum(dr * dr, axis=1)
        m = np.argmax(d2)
        if m in covered:
            return center, radius
        covered.add(m)

        d = np.sqrt(d2[m])
        if d <= radius:
            radius = d
            return center, radius
        p = points[[m]]
        alpha = radius / d
        radius = .5 * (d + radius)
        center = .5 * ((1. - alpha) * p + (1. + alpha) * center)


def infer_spherical_volume(cells, **kwargs):
    """Returns a DataFrame.

    See also `spherical_volume`.
    """
    vol = spherical_volume(cells, 'all', **kwargs)
    return None if vol is None else pd.DataFrame(vol, columns=['volume'])


def spherical_volume(cells, which='all', bounding_sphere='ritter', **kwargs):
    """
    Volume is evaluated looking for a bounding sphere.
    If the sphere does not include locations from neighbour cells as well,
    then its radius is increased until an external location is reached.

    The volume of the cell is approximated by the volume of the resulting sphere.

    For now only the Ritter's bounding sphere algorithm is implemented.

    Returns a Series, or a float if `which` is a single cell index.
    """
    if not callable(bounding_sphere):
        if bounding_sphere.lower() != 'ritter':
            raise ValueError("unsupported bounding sphere algorithm: '{}'".format(bounding_sphere))
        bounding_sphere = ritter
    indices, radii = [], []
    for index in (cells if which == 'all' else which):
        cell = cells[index]
        sphere = bounding_sphere(cell.r, **kwargs)
        if sphere is not None:
            center, radius = sphere
            neighbour_cells = cells.neighbours(index)
            if neighbour_cells.size:
                neighbour_locations = [ cells[neighbour].r for neighbour in neighbour_cells ]
                neighbour_locations = np.vstack(neighbour_locations)
                dist = neighbour_locations - center
                dist = np.sqrt(np.min(np.sum(dist * dist, axis=1)))
                if radius < dist:
                    radius = dist
            else:
                # all the neighbour cells were discarded
                continue
            indices.append(index)
            radii.append(radius)
    if not radii:
        return None
    radii = np.array(radii)
    volumes = 4 / 3 * pi * radii**3

    if which == 'all' or not np.isscalar(which):
        return pd.Series(volumes, index=indices)
    else:
        assert np.isscalar(volumes)
        return volumes.tolist()[0]


__all__ = [ 'infer_spherical_volume', 'spherical_volume', 'setup', 'ritter' ]




# -*- coding: utf-8 -*-

# Copyright © 2017-2018, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from tramway.core import ChainArray
from .base import *
from warnings import warn
from math import pi, log
import numpy as np
import pandas as pd
from scipy.optimize import minimize
from collections import OrderedDict


setup = {'name': ('unsmooth.dd', 'unsmooth.ddrift', 'degraded.dd', 'degraded.ddrift'),
    'provides': ('dd', 'ddrift'),
    'arguments': OrderedDict((
        ('localization_error',  ('-e', dict(type=float, help='localization precision (see also sigma; default is 0.03)'))),
        ('jeffreys_prior',      ('-j', dict(action='store_true', help="Jeffreys' prior"))),
        ('min_diffusivity',     dict(type=float, help='minimum diffusivity value allowed')))),
        'cell_sampling':    'individual'};


def dd_neg_posterior(x, dd, cell, sigma2, jeffreys_prior, dt_mean, min_diffusivity):
    dd.update(x)
    D, drift = dd['D'], dd['drift']
    if D < min_diffusivity:
        warn(DiffusivityWarning(D, min_diffusivity))
    noise_dt = sigma2
    n = len(cell) # number of translocations
    denominator = 4. * (D * cell.dt + noise_dt) # 4*(D+Dnoise)*dt
    dr_minus_drift_dt = cell.dr - np.outer(cell.dt, drift)
    # non-directional squared displacement
    ndsd = np.sum(dr_minus_drift_dt * dr_minus_drift_dt, axis=1)
    neg_posterior = n * log(pi) + np.sum(np.log(denominator)) + np.sum(ndsd / denominator)
    if jeffreys_prior:
        neg_posterior += 2. * log(D * dt_mean + sigma2)
    return neg_posterior


def infer_DD(cells, localization_error=None, jeffreys_prior=False, min_diffusivity=None, **kwargs):
    if isinstance(cells, Distributed): # multiple cells
        localization_error = cells.get_localization_error(kwargs, 0.03, True, \
                localization_error=localization_error)
        args = (localization_error, jeffreys_prior, min_diffusivity)
        index, inferred = [], []
        for i in cells:
            cell = cells[i]
            index.append(i)
            inferred.append(infer_DD(cell, *args, **kwargs))
        any_cell = cell
        inferred = pd.DataFrame(np.stack(inferred, axis=0), \
            index=index, \
            columns=[ 'diffusivity' ] + \
                [ 'drift ' + col for col in any_cell.space_cols ])
        return inferred
    else: # single cell
        cell = cells
        # sanity checks
        if not bool(cell):
            raise ValueError('empty cells')
        if cell.dr.shape[1] == 0:
            raise ValueError('translocation array has no column')
        if cell.dt.shape[1:]:
            raise ValueError('time deltas are structured in multiple dimensions')
        # ensure that translocations are properly oriented in time
        if not np.all(0 < cell.dt):
            warn('translocation dts are non-positive', RuntimeWarning)
            cell.dr[cell.dt < 0] *= -1.
            cell.dt[cell.dt < 0] *= -1.
        #
        dt_mean = np.mean(cell.dt)
        D_initial = np.mean(cell.dr * cell.dr) / (2. * dt_mean)
        initial_drift = np.zeros(cell.dim, dtype=D_initial.dtype)
        dd = ChainArray('D', D_initial, 'drift', initial_drift)
        if min_diffusivity is not False:
            if min_diffusivity is None:
                noise_dt = localization_error
                min_diffusivity = (1e-16 - noise_dt) / np.max(cell.dt)
            kwargs['bounds'] = [(min_diffusivity, None)] + [(None, None)] * cell.dim
        #cell.cache = None # no cache needed
        result = minimize(dd_neg_posterior, dd.combined, \
            args=(dd, cell, localization_error, jeffreys_prior, dt_mean, min_diffusivity), \
            **kwargs)
        #dd.update(result.x)
        #return (dd['D'], dd['drift'])
        return result.x # needless to split dd.combined into D and drift




# -*- coding: utf-8 -*-

# Copyright © 2017, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from ..base import *
from tramway.core.scaler import *
from math import *
import numpy as np
import pandas as pd
from threading import Lock
import itertools
import scipy.sparse as sparse
from scipy.spatial.distance import cdist
from .dichotomy import ConnectedDichotomy
from collections import OrderedDict


def _face_hash(v1, n1):
    '''key that identify a same face.'''
    return tuple(v1) + tuple(n1)


class KDTreeMesh(Voronoi):
    """k-dimensional tree (quad tree in 2D) based tessellation.

    Attributes:
        scaler: see :class:`~tramway.tessellation.base.Tessellation`.
        min_probability (float): minimum probability of a point to be in a given cell.
        max_probability (float): maximum probability of a point to be in a given cell.
        max_level (float): maximum level, considering that the smallest cells are at level 0
            and the level increments each time the cell size doubles.
        _min_distance (float, private): scaled minimum distance between neighbor cell centers.
        _avg_distance (float, private): scaled average distance between neighbor cell centers.
    """
    def __init__(self, scaler=None, min_distance=None, avg_distance=None, \
        min_probability=None, max_probability=None, max_level=None, **kwargs):
        Voronoi.__init__(self, scaler)
        self._min_distance = min_distance
        self._avg_distance = avg_distance
        self.min_probability = min_probability
        if min_probability and not max_probability:
            max_probability = 10.0 * min_probability
        self.max_probability = max_probability
        self.max_level = max_level

    def cell_index(self, points, format=None, select=None, knn=None,
        min_location_count=None, metric='chebyshev', filter=None,
        filter_descriptors_only=False, **kwargs):
        if isinstance(knn, tuple):
            min_nn, max_nn = knn
        else:
            min_nn, max_nn = knn, None
        if format == 'force array':
            min_nn = None
        #if max_nn: # valid only if points are those the tessellation was grown with
        #       max_location_count = int(floor(self.max_probability * points.shape[0]))
        #       if max_nn < max_location_count:
        #           max_nn = None
        if metric == 'chebyshev':
            if min_nn or max_nn:
                raise NotImplementedError('knn support has evolved and KDTreeMesh still lacks a proper support for it. You can still call cell_index with argument metric=\'euclidean\'')
            # TODO: pass relevant kwargs to cdist
            points = self.scaler.scale_point(points, inplace=False)
            X = self.descriptors(points, asarray=True)
            D = cdist(X, self._cell_centers, metric) # , **kwargs
            I, J = np.nonzero(D <= self.reference_length)
            if min_location_count or filter is not None:
                K, count = np.unique(J, return_counts=True)
                if min_location_count:
                    for k in K[count < min_location_count]:
                        J[J == k] = -1
                if filter is not None:
                    if min_location_count:
                        K = K[min_location_count <= K]
                    for k in K:
                        cell = J == k
                        if filter_descriptors_only:
                            x = X[cell]
                        else:
                            x = points[cell]
                        if not filter(self, k, x):
                            J[cell] = -1

            if I[0] == 0 and I.size == points.shape[0] and I[-1] == points.shape[0] - 1:
                K = J
            else:
                K = -np.ones(points.shape[0], dtype=J.dtype)
                K[I] = J
            if format == 'force array':
                format = 'array' # for :func:`format_cell_index`
            return format_cell_index(K, format=format, select=select,
                shape=(points.shape[0], self.cell_adjacency.shape[0]))
        else:
            return Delaunay.cell_index(self, points, format=format, select=select, knn=knn,
                min_location_count=min_location_count, metric=metric, **kwargs)

    def tessellate(self, points, **kwargs):
        init = self.scaler.init
        points = self._preprocess(points)
        if init:
            if self._min_distance is not None:
                self._min_distance = self.scaler.scale_distance(self._min_distance)
            if self._avg_distance is not None:
                self._avg_distance = self.scaler.scale_distance(self._avg_distance)
        min_distance = None
        avg_distance = None
        if self._avg_distance:
            avg_distance = self._avg_distance
        elif self._min_distance:
            min_distance = self._min_distance
        min_count = int(round(self.min_probability * points.shape[0]))
        max_count = int(round(self.max_probability * points.shape[0]))
        self.dichotomy = ConnectedDichotomy(self.descriptors(points, asarray=True), \
            min_count=min_count, max_count=max_count, \
            min_edge=min_distance, base_edge=avg_distance, max_level=self.max_level)
        del self.max_level
        self.dichotomy.split()
        self.dichotomy.subset = {} # clear memory
        self.dichotomy.subset_counter = 0

        origin, level = zip(*[ self.dichotomy.cell[c][:2] for c in range(self.dichotomy.cell_counter) ])
        origin = np.vstack(origin)
        level = np.array(level)
        self._cell_centers = origin + self.dichotomy.reference_length[level[:, np.newaxis] + 1]
        adjacency = np.vstack([ self.dichotomy.adjacency[e] for e in range(self.dichotomy.edge_counter) \
            if e in self.dichotomy.adjacency ]) # not symmetric
        #print(np.sum(np.all(adjacency==np.fliplr(adjacency[0][np.newaxis,:])))) # displays 0
        nedges = adjacency.shape[0]
        self.cell_adjacency = sparse.csr_matrix((np.tile(np.arange(nedges), 2), \
                (adjacency.flatten('F'), np.fliplr(adjacency).flatten('F'))), \
            shape=(self.dichotomy.cell_counter, self.dichotomy.cell_counter))
        # (adjacency[i,0], adjacency[i,1], i)

        # quick and dirty Voronoi construction: vertices are introduced as many times as they
        # appear in a ridge, and as many ridges are introduced as four times the number of
        # centers. In a second step, duplicate vertices are removed.
        def unique_rows(data, *args, **kwargs):
            uniq = np.unique(data.view(data.dtype.descr * data.shape[1]), *args, **kwargs)
            if isinstance(uniq, tuple):
                return (uniq[0].view(data.dtype).reshape(-1, data.shape[1]),) + tuple(uniq[1:])
            else:
                return uniq.view(data.dtype).reshape(-1, data.shape[1])
        n = origin.shape[0]
        vertices = []
        ridge_vertices = []
        for i, v1 in enumerate(self.dichotomy.unit_hypercube):
            vertices.append(origin + \
                np.float_(v1) * self.dichotomy.reference_length[level[:, np.newaxis]])
            for jj, v2 in enumerate(self.dichotomy.unit_hypercube[i+1:]):
                if np.sum(v1 != v2) == 1: # neighbors in the voronoi
                    j = i + 1 + jj
                    ridge_vertices.append(np.vstack(\
                        np.hstack((np.arange(i * n, (i+1) * n)[:,np.newaxis], \
                            np.arange(j * n, (j+1) * n)[:,np.newaxis]))))
        vertices, I = unique_rows(np.concatenate(vertices, axis=0), \
            return_inverse=True)
        self._vertices = vertices
        self._lazy['vertices'] = False
        ridge_vertices = I[np.concatenate(ridge_vertices, axis=0)]
        u, v = ridge_vertices.T
        nverts = vertices.shape[0]
        self.vertex_adjacency = sparse.coo_matrix(\
            (np.ones(2*u.size, dtype=bool), (np.r_[u, v], np.r_[v, u])), \
            shape=(nverts, nverts))
        self.cell_vertices = { i: I[i+n*np.arange(self.dichotomy.unit_hypercube.shape[0])] \
            for i in range(n) }
        # for `cell_index` even after call to `freeze`
        self.reference_length = self.dichotomy.reference_length[level[np.newaxis,:] + 1]
        #self._postprocess()

    def _postprocess(self):
        pass

    def freeze(self):
        self.dichotomy = None

    def delete_cell(self, i, adjacency_label=True, metric='chebyshev', pack_indices=True):
        Voronoi.delete_cell(self, i, adjacency_label, metric, pack_indices)


setup = {
    'make': KDTreeMesh,
    'make_arguments': OrderedDict((
        ('min_distance', ()),
        ('avg_distance', ()),
        ('min_probability', ()),
        ('max_probability', ()),
        ('max_location_count', ('-S', dict(type=int, help='maximum number of locations per cell'))),
        ('max_level', ('-ll', '--lower-levels', dict(type=int, help='number of levels below the smallest one', metavar='LOWER_LEVELS'))),
        )),
    }

__all__ = ['KDTreeMesh', 'setup']




# -*- coding: utf-8 -*-

# Copyright © 2017-2020, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from tramway.core import *
from tramway.tessellation.base import *
import numpy as np
import pandas as pd
import copy
import scipy.sparse as sparse


class TimeLattice(Tessellation):
    """Proxy `Tessellation` for time lattice expansion.

    If `time_lattice` contains integers, these elements are regarded as frame indices,
    whereas if it contains floats, the elements represent time.

    The `time_edge` attribute (`time_label` init argument) drives the encoding of temporal
    adjacency.
    It is a two-element sequence respectively representing *past* and *future* relationships.
    If `time_label` is supplied as a single scalar value, *past* and *future* relationships
    are encoded with this same label.
    If a label is ``False``, then the corresponding relationship is not represented.
    If a label is ``True``, then it is translated into an integer value that is not used yet.

    The `time_dimension` attribute may be useful combined with a defined `spatial_mesh`
    to include time in the representation of cell centers and the calculation of cell volumes.

    If `time_label` is not ``None`` and `time_dimension` is ``None``, `time_dimension` will
    default to ``True``.
    If `time_dimension` is ``True`` and `time_label` is ``None``, then `time_label` will
    default to ``True``.
    ``None`` to both arguments will be treated as ``False``.
    These rules are resolved in :meth:`tessellate`.

    Functional dependencies:

    * setting `time_lattice` unsets `cell_label`, `cell_adjacency` and `adjacency_label`
    * setting `spatial_mesh` unsets `cell_centers`, `cell_label`, `cell_adjacency` and `adjacency_label`, `cell_volume`
    * `cell_centers`, `cell_volume` and `split_frames` are available only when `spatial_mesh`
        is defined

    """
    __slots__ = ('_spatial_mesh', '_time_lattice', 'time_edge', '_cell_centers', '_cell_volume',
        'time_dimension')

    __lazy__ = Tessellation.__lazy__ + \
        ('cell_centers', 'cell_adjacency', 'cell_label', 'adjacency_label', 'cell_volume')

    def __init__(self, scaler=None, segments=None, time_label=None, mesh=None,
            time_dimension=None):
        Tessellation.__init__(self, scaler) # scaler is ignored
        self._time_lattice = segments
        self.time_edge = None if time_label in ((), []) else time_label
        self._cell_adjacency = None
        self._cell_label = None
        self._adjacency_label = None
        self._spatial_mesh = mesh
        self.time_dimension = time_dimension
        self._cell_centers = None
        self._cell_volume = None

    @property
    def time_lattice(self):
        return self._time_lattice

    @time_lattice.setter
    def time_lattice(self, segments):
        self.cell_label = None
        self.cell_adjacency = None
        self.adjacency_label = None
        if isinstance(segments, list):
            segments = np.array(segments)
        # ensure that single segments are encoded as single-row matrices
        if not segments.shape[1:]:
            assert segments.shape[0] == 2
            segments = segments[np.newaxis,:]
        self._time_lattice = segments

    @property
    def spatial_mesh(self):
        return self._spatial_mesh

    @spatial_mesh.setter
    def spatial_mesh(self, tessellation):
        self.cell_label = None
        self.cell_adjacency = None
        self.adjacency_label = None
        self._spatial_mesh = tessellation
        self.cell_centers = None
        self.cell_volume = None

    def tessellate(self, points, time_only=False, **kwargs):
        if self.time_edge in (None, (), []):
            if self.time_dimension is None:
                self.time_edge = None
            else:
                self.time_edge = bool(self.time_dimension)
        elif self.time_dimension is None:
            self.time_dimension = bool(self.time_edge)
        if self.spatial_mesh is not None and not time_only:
            self.spatial_mesh.tessellate(points, **kwargs)

    def cell_index(self, points, *args, **kwargs):
        """
        In addition to the arguments to :meth:`~tramway.tessellation.base.Tessellation.cell_index`
        for the underlying spatial tessellation:

        Arguments:

            time_col (int or str): column name; default is '*t*'; keyword-argument only.

            time_knn (int or tuple): ``min_nn`` minimum number of nearest "neighbours"
                of the time segment center, or ``(min_nn, max_nn)`` minimum and maximum
                numbers of nearest neighbours in time; keyword-argument only.

        `knn` and `time_knn` are not compatible.
        """
        exclude = kwargs.pop('exclude_cells_by_location_count', None)
        # time knn
        time_knn = kwargs.pop('time_knn', None)
        if time_knn is not None:
            knn = kwargs.pop('knn', None)
            if knn is not None:
                raise NotImplementedError('`knn` and `time_knn` are not compatible')
        # extract the timestamps
        time_col = kwargs.pop('time_col', 't')
        if isstructured(points):
            ts = points[time_col]
            if isinstance(ts, (pd.Series, pd.DataFrame)):
                ts = ts.values
        else:
            ts = points[:,time_col]
        time = self.time_lattice
        nsegments = time.shape[0]
        if time.dtype == int:
            t0 = ts.min()
            dt = np.unique(np.diff(np.sort(ts)))
            if dt[0] == 0:
                dt = dt[1]
            else:
                dt = dt[0]
            time = (time * dt) + t0
        #
        if self.spatial_mesh is None:
            #spatial_index = None
            count_shape = (nsegments,)
        else:
            #spatial_index = self.spatial_mesh.cell_index(points, *args, **kwargs)
            ncells = self.spatial_mesh.cell_adjacency.shape[0]
            count_shape = (ncells, nsegments)
        if exclude:
            location_count = np.zeros(count_shape, dtype=int)
        ps, cs = [], []
        if time_knn is None:
            for t in range(nsegments):
                t0, t1 = time[t]
                segment = np.logical_and(t0 <= ts, ts < t1)
                pts, = np.nonzero(segment)
                if pts.size:
                    if self.spatial_mesh is None:
                        ids = np.full_like(pts, t)
                    else:
                        if isinstance(points, pd.DataFrame):
                            points_t = points.iloc[segment]
                        else:
                            points_t = points[segment]
                        ids = self.spatial_mesh.cell_index(points_t, *args, **kwargs)
                        if isinstance(ids, np.ndarray):
                            pass
                        elif isinstance(ids, tuple):
                            _pts, ids = ids
                            pts = pts[_pts]
                        else:
                            raise NotImplementedError
                        if exclude:
                            vs, count = np.unique(ids, return_counts=True)
                            location_count[vs, t] = count
                        ids += t * ncells
                    #if isinstance(points, DataFrame):
                    #       pts = points.index.values[pts] # NO!
                    ps.append(pts)
                    cs.append(ids)
        else:
            if not callable(time_knn):
                try:
                    _min_n, _max_n = time_knn
                except TypeError:
                    _min_n, _max_n = time_knn, None
                else:
                    if _min_n is not None and _max_n is not None and _max_n < _min_n:
                        raise ValueError('time_nn_min > time_nn_max')
            _strict_min_n = kwargs.get('min_location_count', None)
            if _strict_min_n is None:
                _strict_min_n = 0
            if self.spatial_mesh is None:
                for t in range(nsegments):
                    if callable(time_knn):
                        _min_n, _max_n = time_knn(t)
                        # assert _min_n <= _max_n
                    t0, t1 = time[t]
                    segment = np.logical_and(t0 <= ts, ts < t1)
                    _n = np.count_nonzero(segment)
                    if _n < _strict_min_n:
                        continue
                    if _max_n and _n < _max_n:
                        _ts = np.abs(ts[segment] - .5 * (t0 + t1))
                        _i = np.argsort(_ts)
                        _segment = np.zeros(_n, dtype=segment.dtype)
                        _segment[_i[:_max_n]] = True
                        segment[segment] = _segment
                    elif _min_n and _n < _min_n:
                        _ts = np.abs(ts - .5 * (t0 + t1))
                        _i = np.argsort(_ts)
                        segment[_i[:_min_n]] = True
                    pts, = segment.nonzero()
                    if 0 < pts.size:
                        ps.append(pts)
                        cs.append(np.full(pts.shape, t))
            else:
                ids = self.spatial_mesh.cell_index(points, *args, **kwargs)
                if isinstance(ids, np.ndarray):
                    pts = None
                elif isinstance(ids, tuple):
                    pts, ids = ids
                else:
                    raise NotImplementedError
                for i in range(ncells):#np.unique(ids):
                    if pts is None:
                        pts_i, = (ids==i).nonzero()
                    else:
                        pts_i = pts[ids==i]
                    ts_i = ts[pts_i]
                    for t in range(nsegments):
                        if callable(time_knn):
                            _min_n, _max_n = time_knn(i,t) # i= space cell index, t= time segment index
                            # assert _min_n <= _max_n
                        t0, t1 = time[t]
                        segment = np.logical_and(t0 <= ts_i, ts_i < t1)
                        _n = np.count_nonzero(segment)
                        if _n < _strict_min_n:
                            continue
                        if _max_n and _n < _max_n:
                            _ts = np.abs(ts_i[segment] - .5 * (t0 + t1))
                            _i = np.argsort(_ts)
                            _segment = np.zeros(_n, dtype=segment.dtype)
                            _segment[_i[:_max_n]] = True
                            segment[segment] = _segment
                        elif _min_n and _n < _min_n:
                            _ts = np.abs(ts_i - .5 * (t0 + t1))
                            _i = np.argsort(_ts)
                            segment[_i[:_min_n]] = True
                        pts_t, = segment.nonzero()
                        if 0 < pts_t.size:
                            pts_t = pts_i[pts_t]
                            ps.append(pts_t)
                            cs.append(np.full(pts_t.shape, t * ncells + i, dtype=int)) # `dtype=int` is to turn a numpy warning off and optimize for old numpy versions
        if ps:
            if exclude and not count_shape[1:]:
                ok = exclude(location_count)
                ok = ok[0 < location_count]
                if not np.any(ok):
                    return ([], [])
                ps, cs = zip(*[ (p, c) for p, c, b in zip(ps, cs, ok) if b ])
            ps = np.concatenate(ps)
            cs = np.concatenate(cs)
            if exclude and count_shape[1:]:
                i, t = exclude(location_count).nonzero()
                ok = np.ones(cs.size, dtype=bool)
                for c in t * ncells + i:
                    ok[cs == c] = False
                ps = ps[ok]
                cs = cs[ok]
        return (ps, cs)

    # descriptors
    def descriptors(self, points, *args, **kwargs):
        if self.spatial_mesh is None:
            if args:
                asarray = args.pop(0)
            else:
                asarray = kwargs.pop('asarray', False)
            time_col = kwargs.pop('time_col', 't')
            if isstructured(points):
                timestamps = points[time_col]
            else:
                timestamps = points[:,time_col]
            if asarray:
                return np.asarray(timestamps)
            else:
                return timestamps
        else:
            return self.spatial_mesh.descriptors(points, *args, **kwargs)

    # cell_adjacency property
    @property
    def cell_adjacency(self):
        if self._cell_adjacency is None:
            nsegments = self.time_lattice.shape[0]

            if self.time_edge in (None, (), []):
                self.time_edge = past_edge = future_edge = None
            else:
                try:
                    past_edge, future_edge = self.time_edge
                except (TypeError, ValueError):
                    past_edge = future_edge = self.time_edge
                if past_edge is False:
                    past_edge = None
                if future_edge is False:
                    future_edge = None

            if self.spatial_mesh is None:
                cell_ids = np.arange(nsegments)
                row, col, data = [], [], []

                # the first label in `labels` is fake (at index 0)
                if past_edge is True:
                    if future_edge is True:
                        past_edge, future_edge = -1, 1
                    elif future_edge is None:
                        past_edge = -1
                    else:
                        past_edge = future_edge - 1
                    labels = [past_edge - 1]
                elif future_edge is True:
                    if past_edge is None:
                        future_edge = 1
                    else:
                        future_edge = past_edge + 1
                    labels = [future_edge + 1]
                else:
                    labels = [min(past_edge, future_edge) - 1]

                if past_edge is not None:
                    row.append(cell_ids[:-1])
                    col.append(cell_ids[1:])
                    data.append(np.full(nsegments - 1, len(labels), dtype=int))
                    labels.append(past_edge)

                if future_edge is not None:
                    row.append(cell_ids[1:])
                    col.append(cell_ids[:-1])
                    data.append(np.full(nsegments - 1, len(labels), dtype=int))
                    labels.append(future_edge)

                if data:
                    data = np.concatenate(data)
                    row = np.concatenate(row)
                    col = np.concatenate(col)
                    self._cell_adjacency = sparse.coo_matrix((data, (row, col)),
                        shape=(nsegments, nsegments))
                    self._adjacency_label = np.array(labels)
                else:
                    self._cell_adjacency = sparse.coo_matrix((nsegments, nsegments),
                        dtype=bool)
                    self._adjacency_label = []

            else:
                if self.spatial_mesh.adjacency_label is None:
                    A = sparse.triu(self.spatial_mesh.cell_adjacency, format='coo')
                    edge_max = int(A.data.max())
                    if 1 < edge_max:
                        raise ValueError('non-boolean values in the adjacency matrix are not indices of labels or the labels are missing')
                    n_spatial_edges = A.data.size
                    A = sparse.coo_matrix((np.tile(np.arange(n_spatial_edges), 2), \
                            (np.r_[A.row, A.col], np.r_[A.col, A.row])), \
                        shape=A.shape).tocsr()
                    self._adjacency_label = np.ones(n_spatial_edges, dtype=int)
                else:
                    self._adjacency_label = self.spatial_mesh.adjacency_label
                    A = self.spatial_mesh.cell_adjacency.tocsr()
                    edge_max = int(A.data.max())
                    if edge_max + 1 < self._adjacency_label.size:
                        self._adjacency_label = self._adjacency_label[:edge_max+1]

                ncells = A.shape[0]
                active_cells, = np.where(0 < np.diff(A.indptr))
                edge_ptr = self._adjacency_label.size

                if past_edge is None:
                    past = None
                else:
                    past = sparse.coo_matrix( \
                        (np.arange(edge_ptr, edge_ptr + active_cells.size), \
                            (active_cells, active_cells)), \
                        shape=(ncells, ncells))
                    edge_ptr += active_cells.size
                if future_edge is None:
                    future = None
                else:
                    future = sparse.coo_matrix( \
                        (np.arange(edge_ptr, edge_ptr + active_cells.size), \
                            (active_cells, active_cells)), \
                        shape=(ncells, ncells))
                    edge_ptr += active_cells.size

                if nsegments == 1:
                    self._cell_adjacency = A
                else:
                    blocks = [[A, future] + [None] * (nsegments - 2)]
                    for k in range(1, nsegments - 1):
                        blocks.append([None] * (k - 1) + [past, A, future] + \
                            [None] * (nsegments - 2 - k))
                    blocks.append([None] * (nsegments - 2) + [past, A])
                    self._cell_adjacency = sparse.bmat(blocks, format='csr')

                if past_edge is True:
                    if future_edge != edge_max + 1:
                        past_edge = edge_max + 1
                    else:
                        past_edge = max(edge_max, future_edge) + 1
                    edge_max += 1
                if future_edge is True:
                    if past_edge is None:
                        future_edge = edge_max + 1
                    else:
                        future_edge = max(edge_max, past_edge) + 1
                    edge_max += 1
                dtype = self._adjacency_label.dtype
                if past_edge and future_edge:
                    self._adjacency_label = np.r_[self._adjacency_label, \
                        np.full(active_cells.size, past_edge, dtype=dtype), \
                        np.full(active_cells.size, future_edge, dtype=dtype)]
                elif past_edge:
                    self._adjacency_label = np.r_[self._adjacency_label, \
                        np.full(active_cells.size, past_edge, dtype=dtype)]
                elif future_edge:
                    self._adjacency_label = np.r_[self._adjacency_label, \
                        np.full(active_cells.size, future_edge, dtype=dtype)]

            self.time_edge = (past_edge, future_edge)
        return self.__returnlazy__('cell_adjacency', self._cell_adjacency)

    @cell_adjacency.setter
    def cell_adjacency(self, matrix):
        self.__lazysetter__(matrix)

    # past/future properties
    @property
    def past_edge(self):
        return self.time_edge[0]

    @property
    def future_edge(self):
        return self.time_edge[1]

    # cell_label
    @property
    def cell_label(self):
        if self._cell_label is None:
            if self.spatial_mesh is None or self.spatial_mesh.cell_label is None:
                return None
            else:
                nsegments = self.time_lattice.shape[0]
                return np.tile(self.spatial_mesh.cell_label, nsegments)
        else:
            return self.__returnlazy__('cell_label', self._cell_label)

    @cell_label.setter
    def cell_label(self, label):
        self.__lazysetter__(label)

    # adjacency_label
    @property
    def adjacency_label(self):
        if self._adjacency_label is None:
            self.cell_adjacency
        return self.__returnlazy__('adjacency_label', self._adjacency_label)

    @adjacency_label.setter
    def adjacency_label(self, label):
        self.__lazysetter__(label)

    def simplified_adjacency(self, adjacency=None, label=None, format='coo',
            distinguish_time=None):
        """
        `distinguish_time` allows to keep temporal relationships distinct from the spatial
        relationships.
        As a consequence of encoding time, the simplified adjacency matrix is not boolean.

        If `distinguish_time` is ``None``, then `distinguish_time` will default to ``True``
        if `time_edge` is defined, ``False`` otherwise.
        """
        if distinguish_time is False or \
            (distinguish_time is None and self.time_edge == (None, None)):
            return Tessellation.simplified_adjacency(self, adjacency, label, format)
        if adjacency is None:
            adjacency = self.cell_adjacency
        #else: beware that self.adjacency_label is used anyway
        _adjacency = Tessellation.simplified_adjacency(self, adjacency, label, 'coo')
        # cannot squeeze adjacency[_adjacency.row, _adjacency.col]...
        _labels = self.adjacency_label[adjacency[_adjacency.row, _adjacency.col]]
        _adjacency.data = _adjacency.data.astype(int)
        _i = 1
        for _label in self.time_edge:
            _i += 1
            _adjacency.data[(_labels == _label).nonzero()[1]] = _i
        if format == 'csr':
            _adjacency = _adjacency.tocsr()
        elif format == 'csc':
            _adjacency = _adjacency.tocsc()
        elif format == 'lil':
            _adjacency = _adjacency.tolil()
        elif format == 'dok':
            _adjacency = _adjacency.todok()
        elif format == 'dia':
            _adjacency = _adjacency.todia()
        elif format == 'bsr':
            _adjacency = _adjacency.tobsr()
        else:
            raise NotImplementedError('unsupported sparse matrix format')
        return _adjacency

    @property
    def number_of_cells(self):
        n_segments = self.time_lattice.shape[0]
        if self.spatial_mesh is None:
            return n_segments
        else:
            return n_segments * self.spatial_mesh.number_of_cells

    ## Delaunay properties and methods
    @property
    def cell_centers(self):
        if self._cell_centers is None:
            if self.time_dimension and self.time_lattice.dtype == int:
                raise ValueError('time is encoded as frame indices')
            nsegments = self.time_lattice.shape[0]
            if self.spatial_mesh is None:
                raise AttributeError('`cell_centers` is defined only for time lattices combined with spatial tessellations')
                self._cell_centers = np.mean(self.time_lattice, axis=1) # valid only if `time_lattice` are timestamps!
            else:
                self._cell_centers = self.spatial_mesh.cell_centers
                ncells = self._cell_centers.shape[0]
                self._cell_centers = np.tile(self._cell_centers, (nsegments, 1))
                if self.time_dimension:
                    self._cell_centers = np.hstack((self._cell_centers, \
                        np.repeat(np.mean(self.time_lattice, axis=1), \
                            ncells)[:,np.newaxis]))
        return self.__returnlazy__('cell_centers', self._cell_centers)

    @cell_centers.setter
    def cell_centers(self, pts):
        if pts is None:
            self.__lazysetter__(pts)
        elif self.spatial_mesh is None:
            raise AttributeError('`cell_centers` is defined only for time lattices combined with spatial tessellations')
        else:
            raise AttributeError('`cell_centers` is read-only')


    # Voronoi properties and methods
    @property
    def cell_volume(self):
        """
        If `time_dimension` is ``True``, then the product of spatial volume and
        segment duration is returned.
        Otherwise, `cell_volume` is only the spatial volume.
        """
        if self._cell_volume is None:
            if self.time_dimension and self.time_lattice.dtype == int:
                raise ValueError('time is encoded as frame indices')
            nsegments = self.time_lattice.shape[0]
            if self.spatial_mesh is None:
                raise AttributeError('`cell_volume` is defined only for time lattices combined with spatial tessellations')
                self._cell_volume = np.diff(self.time_lattice, axis=1)
            else:
                self._cell_volume = self.spatial_mesh.cell_volume
                ncells = self._cell_volume.shape[0]
                self._cell_volume = np.tile(self._cell_volume, nsegments)
                if self.time_dimension:
                    segment_duration = np.squeeze(np.diff(self.time_lattice, axis=1))
                    self._cell_volume *= \
                        np.repeat(segment_duration, ncells)
        return self.__returnlazy__('cell_volume', self._cell_volume)

    @cell_volume.setter
    def cell_volume(self, v):
        if v is None:
            self.__lazysetter__(v)
        elif self.spatial_mesh is None:
            raise AttributeError('`cell_volume` is defined only for time lattices combined with spatial tessellations')
        else:
            raise AttributeError('`cell_volume` is read-only')


    # other methods
    def split_frames(self, df, return_times=False):
        if not isinstance(df, (pd.Series, pd.DataFrame)):
            raise TypeError('implemented only for `pandas.DataFrame`s')
        if self.spatial_mesh is None:
            raise NotImplementedError('missing spatial tessellation')
        ncells = self.spatial_mesh.cell_adjacency.shape[0]
        nsegments = self.time_lattice.shape[0]
        try:
            # not tested yet
            segment, cell = np.divmod(df.index, ncells) # 1.13.0 <= numpy
        except AttributeError:
            try:
                segment = df.index // ncells
            except TypeError:
                print(df.index)
                raise
            cell = np.mod(df.index, ncells)
        ts = []
        for t in range(nsegments):
            xt = df[segment == t]
            xt.index = cell[segment == t]
            if return_times:
                if self.time_lattice.dtype == int:
                    raise ValueError('cannot return timestamps')
                ts.append((self.time_lattice[t], xt))
            else:
                ts.append(xt)
        return ts

    def split_segments(self, spt_data, return_times=False):
        if self.spatial_mesh is None:
            #raise NotImplementedError('missing spatial tessellation')
            ncells = 1
        else:
            ncells = self.spatial_mesh.cell_adjacency.shape[0]
        nsegments = self.time_lattice.shape[0]
        ts = []
        try:
            spt_data = spt_data.maps
        except AttributeError:
            pass
        if isinstance(spt_data, (pd.Series, pd.DataFrame)):
            df = spt_data
            if ncells == 1:
                segment = np.asarray(df.index)
                cell = np.zeros_like(segment)
            else:
                # borrowed from `split_frames`
                try:
                    segment, cell = np.divmod(df.index, ncells) # 1.13.0 <= numpy
                except AttributeError:
                    try:
                        segment = df.index // ncells
                    except TypeError:
                        print(df.index)
                        raise
                    cell = np.mod(df.index, ncells)
            for t in range(nsegments):
                xt = df[segment == t]
                xt.index = cell[segment == t]
                if return_times:
                    if self.time_lattice.dtype == int:
                        raise ValueError('cannot return timestamps')
                    ts.append((self.time_lattice[t], xt))
                else:
                    ts.append(xt)
            #
        elif spt_data.tessellation is self:
            df = spt_data.points
            tessellation = self.spatial_mesh
            for t in range(nsegments):
                p,c = format_cell_index(spt_data.cell_index, 'tuple')
                seg = (t*ncells<=c) & (c<(t+1)*ncells)
                if seg.size == 0:
                    import warnings
                    warnings.warn('empty segment; please check the implementation')
                seg_p = np.unique(p[seg])
                points = df.iloc[seg_p]
                wrong = p.max()+1
                p_tr = np.full(p.max()+1, wrong, dtype=int)
                p_tr[seg_p] = np.arange(seg_p.size)
                seg_p = p_tr[p[seg]]
                assert not np.any(seg_p == wrong)
                seg_c = c[seg] - t*ncells
                assignment = (seg_p, seg_c)
                #
                cells = type(spt_data)(points, tessellation, assignment)
                cells.param = dict(spt_data.param)
                cells.bounding_box = spt_data.bounding_box
                #
                if return_times:
                    ts.append((self.time_lattice[t], cells))
                else:
                    ts.append(cells)
        else:
            raise ValueError('unsupported input argument')
        return ts

    def freeze(self):
        if self.spatial_mesh is not None:
            self.spatial_mesh.freeze()


def with_time_lattice(cells, frames, exclude_cells_by_location_count=None, **kwargs):
    dynamic_cells = copy.deepcopy(cells)
    dynamic_cells.tessellation = TimeLattice(mesh=cells.tessellation, segments=frames)
    dynamic_cells.cell_index = dynamic_cells.tessellation.cell_index(cells.points, \
        exclude_cells_by_location_count=exclude_cells_by_location_count, **kwargs)
    return dynamic_cells


__all__ = ['TimeLattice', 'with_time_lattice']




# -*- coding: utf-8 -*-

# Copyright © 2017, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from .base import *
from .time import *
from .nesting import *
from tramway.core.plugin import Plugins
import os.path


try:
    from .grid import RegularMesh
except ImportError:
    pass


plugins = Plugins(os.path.dirname(__file__), __package__, \
    {'make': Tessellation}, require='setup')




# -*- coding: utf-8 -*-

# Copyright © 2017, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from math import *
import numpy as np
import pandas as pd
import scipy.sparse as sparse
import scipy.stats as stats
from ..base import *
from tramway.core.scaler import *
from .gas import Gas
from scipy.spatial.distance import cdist
import time
from collections import OrderedDict
from warnings import warn


class GasMesh(Voronoi):
    """GWR based tessellation.

    Attributes:
        gas (:class:`~tramway.tessellation.gwr.gas.Gas`):
            internal graph representation of the gas.
        min_probability (float):
            minimum probability of a point to be in any given cell.
        _min_distance (float):
            scaled minimum distance between adjacent cell centers.
        _avg_distance (float):
            upper bound on the average scaled distance between adjacent cell centers.
        _max_distance (float):
            scaled maximum distance between adjacent cell centers."""
    def __init__(self, scaler=None, min_distance=None, avg_distance=None, max_distance=None, \
        min_probability=None, avg_probability=None, **kwargs):
        Voronoi.__init__(self, scaler)
        self.gas = None
        self._min_distance = min_distance
        if avg_distance or max_distance is None:
            self._avg_distance = avg_distance
        else:
            self._avg_distance = max_distance * 0.25
        if max_distance or avg_distance is None:
            self._max_distance = max_distance
        else:
            self._max_distance = avg_distance * 4
        self.min_probability = min_probability
        #self.avg_probability = avg_probability

    def _preprocess(self, points, batch_size=10000, tau=333.0, trust=1.0, lifetime=50, **kwargs):
        if isinstance(points, tuple):
            points, displacements = points
        else:
            displacements = None
        init = self.scaler.init
        points = Voronoi._preprocess(self, points)
        if init:
            if self._min_distance is not None:
                self._min_distance = self.scaler.scale_distance(self._min_distance)
            if self._avg_distance is not None:
                if self._avg_distance <= 0:
                    raise ValueError('`avg_distance` is null or negative')
                self._avg_distance = self.scaler.scale_distance(self._avg_distance)
            if self._max_distance is not None:
                if self._max_distance <= 0:
                    raise ValueError('`max_distance` is null or negative')
                self._max_distance = self.scaler.scale_distance(self._max_distance)
        if self.gas is None:
            self.gas = Gas(np.asarray(points))
            if self._max_distance:
                # distances are diameters, while insertion thresholds should be radii
                self.gas.insertion_threshold = (self._avg_distance * 0.5, \
                    self._max_distance * 0.5)
                if self.min_probability:
                    self.gas.knn = int(round(self.min_probability * \
                        points.shape[0]))
                    if self.gas.knn == 0:
                        if self.min_probability<=1e-8:
                            warn('rounding min_probability down to 0', RuntimeWarning)
                            self.min_probability = 0
                            self.gas.knn = 20
                        else:
                            warn('min_probability is too low for the amount of available data', RuntimeWarning)
                            self.gas.knn = 1
                else:
                    self.gas.knn = 20
            self.gas.trust = trust
            self.gas.batch_size = batch_size
            if not isinstance(tau, tuple):
                tau = (tau, tau)
            self.gas.habituation_tau = tau
            self.gas.edge_lifetime = lifetime
            if self._min_distance:
                self.gas.collapse_below = self._min_distance# * 0.9
        return points, displacements

    def tessellate(self, points, pass_count=(), residual_factor=.7, error_count_tol=5e-3, \
        min_growth=1e-4, collapse_tol=.01, stopping_criterion=0, verbose=False, \
        plot=False, alpha_risk=1e-15, grab=None, max_frames=None, max_batches=None, axes=None, \
        complete_delaunay=False, topology='approximate density', \
        **kwargs):
        """Grow the tessellation.

        Arguments:
            points: see :meth:`~tramway.tessellation.Tessellation.tessellate`.
            pass_count (float or pair of floats):
                number of points to sample (with replacement) from data `points`, as
                a multiple of the size of the data.
                If `pass_count` is a pair of numbers, they are the lower and the upper
                bounds on the number of samples.
                If `pass_count` is a single number, it is interpreted as the lower
                bound, and the upper bound is set equal to ``2 * pass_count``.
            residual_factor (float): multiplies with `_max_distance` to determine
                `residual_max` in :meth:`~tramway.tessellation.gwr.gas.Gas.train`.
            error_count_tol (float): (see :meth:`~tramway.tessellation.gwr.gas.Gas.train`)
            min_growth (float): (see :meth:`~tramway.tessellation.gwr.gas.Gas.train`)
            collapse_tol (float): (see :meth:`~tramway.tessellation.gwr.gas.Gas.train`)
            stopping_criterion (int): (see :meth:`~tramway.tessellation.gwr.gas.Gas.train`)
            verbose (bool): verbose output.
            batch_size (int): (see :class:`~tramway.tessellation.gwr.gas.Gas`)
            tau (float): (see :class:`~tramway.tessellation.gwr.gas.Gas`)
            trust (float): (see :class:`~tramway.tessellation.gwr.gas.Gas`)
            lifetime (int): (see :class:`~tramway.tessellation.gwr.gas.Gas`)
            alpha_risk (float): location distributions of potential neighbor cells
                are compared with a t-test
            complete_delaunay (bool): complete the Delaunay graph
            topology (str): any of 'approximate density' (default), 'displacement length'

        Returns:
            See :meth:`~tramway.tessellation.Tessellation.tessellate`.

        See also:
            :class:`tramway.tessellation.gwr.gas.Gas` and
            :meth:`tramway.tessellation.gwr.gas.Gas.train`.
        """
        #np.random.seed(15894754) # to benchmark and compare between Graph implementations
        points, displacements = self._preprocess(points, **kwargs)
        if self._avg_distance:
            residual_factor *= self._avg_distance # important: do this after _preprocess!
        if pass_count is not None:
            if pass_count == (): # () has been chosen to denote default (not-None) value
                n = points.shape[0]
                p = .95
                # sample size for each point to have probability p of being chosen once
                pass_count = log(1.0 - p) / log(1.0 - 1.0 / float(n))
                # convert in number of passes
                pass_count /= float(n)
            try:
                len(pass_count) # sequence?
            except TypeError:
                pass_count = (pass_count, 2 * pass_count)
        if not (topology is None or topology == 'approximate density'):
            self.gas.topology = topology
        self.residuals = self.gas.train( \
            np.asarray(points) if displacements is None else (np.asarray(points), np.asarray(displacements)), \
            pass_count=pass_count, \
            residual_max=residual_factor, \
            error_count_tol=error_count_tol, \
            min_growth=min_growth, \
            collapse_tol=collapse_tol, \
            stopping_criterion=stopping_criterion, \
            verbose=verbose, plot=plot, \
            grab=grab, max_frames=max_frames, max_batches=max_batches, axes=axes)
        # build alternative representation of the gas (or Delaunay graph)
        [self._cell_adjacency, V, _] = self.gas.export()
        self._cell_centers = V['weight']
        self._cell_adjacency.data = np.ones_like(self._cell_adjacency.data, dtype=int)
        self.alpha_risk = alpha_risk
        self._postprocess(points, complete_delaunay, verbose, _update_cell_adjacency=True)

    def _postprocess(self, points=None, complete_delaunay=False, verbose=False, _update_cell_adjacency=False):
        # build the Voronoi graph
        voronoi = Voronoi._postprocess(self)
        if not _update_cell_adjacency:
            return voronoi # stop here

        # clean and extend the adjacency matrix with the Delaunay graph
        adjacency = self._cell_adjacency # shorter name
        # fix for issue on reload
        if 4 < adjacency.data[-1]:
            adjacency.data[:] = 1
        delaunay = sparse.csr_matrix( \
            (np.ones(2*voronoi.ridge_points.shape[0], dtype=int), \
            (voronoi.ridge_points.flatten('F'), \
            np.fliplr(voronoi.ridge_points).flatten('F'))), \
            shape=adjacency.shape)
        # if `ridge_points` includes the same ridge twice,
        # the corresponding elements in `data` are added
        if not np.all(delaunay.data == 1):
            warn('some Voronoi ridges appear twice', RuntimeWarning)
        delaunay.data[:] = 2
        A = sparse.tril(adjacency + delaunay, format='coo')
        assert delaunay.data.size/2 <= A.data.size
        self._adjacency_label = A.data # labels are: 1=gas only, 2=voronoi only, 3=both
        # edge indices for _adjacency_label
        adjacency = sparse.csr_matrix( \
            (np.tile(np.arange(0, self._adjacency_label.size), 2), \
            (np.concatenate((A.row, A.col)), np.concatenate((A.col, A.row)))), \
            shape=adjacency.shape)
        self.cell_adjacency = adjacency

        ## reintroduce Delaunay edges that do not appear in the gas
        # (turn some label-2 edges into label-4)

        if complete_delaunay:
            self._adjacency_label[self._adjacency_label==2] = 4

        elif points is not None:
            #t = time.time()
            points = np.asarray(points)
            try:
                ix = np.argmin(cdist(points, self._cell_centers), axis=1)
            except MemoryError:
                X, Y = points, self._cell_centers
                # slice X to process less rows at a time (borrowed from tessellation.base.Delaunay.cell_index)
                ix = np.zeros(X.shape[0], dtype=int)
                X2 = np.sum(X * X, axis=1, keepdims=True).astype(np.float32)
                Y2 = np.sum(Y * Y, axis=1, keepdims=True).astype(np.float32)
                X, Y = X.astype(np.float32), Y.astype(np.float32)
                n = 0
                while True:
                    n += 1
                    block = int(ceil(X.shape[0] * 2**(-n)))
                    try:
                        np.empty((block, Y.shape[0]), dtype=X.dtype)
                    except MemoryError:
                        pass # continue
                    else:
                        break
                n += 2 # safer
                block = int(ceil(X.shape[0] * 2**(-n)))
                for i in range(0, X.shape[0], block):
                    j = min(i+block, X2.size)
                    Di = np.dot(np.float32(-2.)* X[i:j], Y.T)
                    Di += X2[i:j]
                    Di += Y2.T
                    ix[i:j] = np.argmin(Di, axis=1)
            #
            ref = int( ceil(float(self.gas.knn) / 8.0) ) # int and float for PY2
            #ref = -ref # with alternative to cdist, index from the end
            ref -= 1 # with cdist, index
            A = sparse.tril(self._cell_adjacency, format='coo') # in future scipy version, check that tril does not remove explicit zeros
            assert np.any(A.data == 0)
            # compute the median distance between adjacent cell centers
            pts_i = np.stack([ self._cell_centers[i] for i in A.row ])
            pts_j = np.stack([ self._cell_centers[j] for j in A.col ])
            ref_d = np.sqrt(np.median(np.sum((pts_i - pts_j)**2, axis=1)))
            for i, j, k in zip(A.row, A.col, A.data):
                if self._adjacency_label[k] == 2: # only in Voronoi
                    xi = points[ix == i]
                    xj = points[ix == j]
                    if 1 < xi.shape[0] and 1 < xj.shape[0]:
                        dij = cdist(xi, xj)
                        #dij = np.dot(xi, xj.T)
                        #xi2 = np.sum(xi * xi, axis=1, keepdims=True)
                        #dij -= 0.5 * xi2
                        #xj2 = np.sum(xj * xj, axis=1, keepdims=True)
                        #dij -= 0.5 * xj2.T
                        dij = dij.flatten()
                        #kij = dij.argsort()
                        dij.sort()
                        if ref_d * .9 < dij[0]:
                            continue
                        try:
                            dij = dij[ref]
                        except IndexError:
                            if verbose and 1 < verbose:
                                print('skipping edge {:d} between cell {:d} (card = {:d}) and cell {:d} (card = {:d}): number of between-cell pairs = {:d} (expected: {:d})'.format(k, i, xi.shape[0], j, xj.shape[0], dij.size, ref))
                            continue
                        #dij = np.sqrt(-2.0 * dij)
                        if dij < self._min_distance:
                            self._adjacency_label[k] = 4 # mark edge as 'not congruent but valid'
                            continue
                        ci, cj = np.mean(xi, axis=0), np.mean(xj, axis=0)
                        yi = np.sort(np.dot(xi - ci, cj - ci))
                        yj = np.sort(np.dot(xj - ci, cj - ci))
                        # throttle the number of points down to control the p-value
                        n0 = 10
                        yi = yi[::-1][:min(max(n0, (ref+1)*2), yi.size)]
                        yj = yj[:min(max(n0, (ref+1)*2), yj.size)]
                        t, p = stats.ttest_ind(yi.T, yj.T, equal_var=False)
                        if self.alpha_risk < p:
                            self._adjacency_label[k] = 4 # mark edge as 'not congruent but valid'
                        continue
                        # debug candidate edges
                        # comment out the above `continue` statement
                        # and uncomment the `argsort` line
                        cell_i_k, cell_j_k = np.unravel_index(kij[ref], (xi.shape[0], xj.shape[0]))
                        if not hasattr(self, 'candidate_edges'):
                            self.candidate_edges = {}
                        self.candidate_edges[k] = (xi[cell_i_k], xj[cell_j_k])
                        #
                    elif verbose and 1 < verbose:
                        print('skipping edge {:d} between cell {:d} (card = {:d}) and cell {:d} (card = {:d})'.format(k, i, xi.shape[0], j, xj.shape[0]))
            sparsity = np.float(np.sum(self._adjacency_label==2)) / np.float(np.sum(0<self._adjacency_label))
            if .5 < sparsity:
                warn('the Delaunay-like graph is very sparse compared to the actual Delaunay graph; pass `complete_delaunay=True` to get the Delaunay graph instead', RuntimeWarning)

        new_labels = np.array([0,-1,-2,1,2])
        # before: 0=[none], 1=not congruent (gas only), 2=not congruent (voronoi only),
        #     3=congruent, 4=congruent after post-processing (initially voronoi only)
        # after: -2=not congruent (voronoi only), -1=not congruent (gas only), 0=[none],
        #     1=congruent, 2=congruent after post-processing (initially voronoi only)
        self._adjacency_label = new_labels[self._adjacency_label]

        return voronoi

    def freeze(self):
        self.gas = None



setup = {
    'name': ('gas', 'gwr'),
    'make': GasMesh,
    'make_arguments': OrderedDict((
        ('min_distance', ()),
        ('avg_distance', ()),
        ('max_distance', ()),
        ('min_probability', ()),
        ('pass_count', dict(type=float, help='fraction of the data to be sampled; can be greater than 1 (recommended)')),
        )),
    }

__all__ = ['GasMesh', 'setup']





from .time import TimeLattice
from ..core import isstructured, hdf5
from collections import OrderedDict
import numpy as np
import pandas as pd


__all__ = ['setup', 'SlidingWindow']


setup = {
    'make_arguments': OrderedDict((
        ('duration', dict(type=float, help="window width in seconds (or in frames)")),
        ('shift', dict(type=float, help="time shift between consecutive segments, in seconds (or in frames)")),
        ('frames', dict(action='store_true', help="regard the --duration and --shift arguments as numbers of frames instead of timestamps")),
        )),
    'window_compatible': False,
    }


class SlidingWindow(TimeLattice):

    __slots__ = ('duration', 'shift', 'start_time')

    def __init__(self, scaler=None, duration=None, shift=None, frames=False, time_label=None,
        time_dimension=None, start_time=None):
        TimeLattice.__init__(self, scaler, time_label=time_label, time_dimension=time_dimension)
        if duration is None:
            raise ValueError("'duration' is required")
        elif np.isclose(max(0, duration), 0):
            raise ValueError("'duration' is too short")
        if shift is None:
            shift = duration
        elif np.isclose(max(0, shift), 0):
            raise ValueError("'shift' is too small")
        if frames:
            duration = int(duration)
            shift = int(shift)
        else:
            duration = float(duration)
            shift = float(shift)
        self.duration = duration
        self.shift = shift
        self.start_time = start_time

    def cell_index(self, points, *args, **kwargs):
        if self.time_lattice is None:
            time_col = kwargs.get('time_col', 't')
            if isstructured(points):
                ts = points[time_col]
                if isinstance(ts, (pd.Series, pd.DataFrame)):
                    ts = ts.values
            else:
                ts = points[:,time_col]
            t0, t1 = ts.min(), ts.max()
            if self.start_time is not None:
                t0 = self.start_time
            duration, shift = self.duration, self.shift
            if isinstance(duration, int):
                dt = np.median(np.diff(np.unique(ts)))
                duration *= dt
                shift *= dt
                dt /= 10.
            else:
                dt = 1e-7 # precision down to a microsecond (quantum < microsecond)
            nsegments = max(1., np.round((t1 - t0 - duration) / shift) + 1.)
            t1 = t0 + (nsegments - 1.) * shift + duration
            t0s = np.arange(t0, t1 - duration + dt, shift)
            t1s = t0s + duration
            self.time_lattice = np.stack((t0s, t1s), axis=-1)
        return TimeLattice.cell_index(self, points, *args, **kwargs)


import sys
if sys.version_info[0] < 3:

    import rwa
    sliding_window_exposes = hdf5.time_lattice_exposes + list(SlidingWindow.__slots__)#['duration', 'shift']
    rwa.hdf5_storable(rwa.default_storable(SlidingWindow, exposes=sliding_window_exposes),
        agnostic=True)

    __all__.append('sliding_window_exposes')




# -*- coding: utf-8 -*-

# Copyright © 2017-2019, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import rwa.lazy as rwa
from . import base
from . import abc


class InstancesView(base.InstancesView):

    __slots__ = ('peek', )

    def __init__(self, analyses, peek=False):
        base.InstancesView.__init__(self, analyses)
        self.peek = peek

    def __getitem__(self, label):
        instance = base.InstancesView.__getitem__(self, label)
        if rwa.islazy(instance):
            if issubclass(instance.type, abc.Analyses):
                instance = instance.shallow()
                self.__setitem__(label, instance)
            elif self.peek:
                instance = instance.deep()
                self.__setitem__(label, instance)
        return instance

    def get(self, label, default=None):
        instance = base.InstancesView.get(self, label, None)
        if instance is None:
            instance = default
        elif rwa.islazy(instance):
            if issubclass(instance.type, abc.Analyses):
                instance = instance.shallow()
                self.__setitem__(label, instance)
            elif self.peek:
                instance = instance.deep()
                self.__setitem__(label, instance)
        return instance

    def pop(self, label, default=None):
        return rwa.lazyvalue(base.InstancesView.pop(self, label, default), deep=self.peek)


class Analyses(base.Analyses):
    __slots__ = ()

    @property
    def type(self):
        return rwa.lazytype(self._data)

    @property
    def data(self):
        if rwa.islazy(self._data):
            self._data = self._data.peek(deep=True)
        return self._data

    # copy/paste
    @data.setter
    def data(self, d):
        self._data = d
        self._instances = {}
        self._comments = {}
        self._metadata = {}

    @property
    def comments(self):
        if rwa.islazy(self._comments):
            self._comments = self._comments.deep()
            if rwa.islazy(self._metadata):
                self._metadata = self._metadata.deep()
            if rwa.islazy(self._instances):
                self._instances = self._instances.shallow()
        return base.CommentsView(self)

    @property
    def metadata(self):
        if rwa.islazy(self._metadata):
            self._metadata = self._metadata.deep()
            if rwa.islazy(self._comments):
                self._comments = self._comments.deep()
            if rwa.islazy(self._instances):
                self._instances = self._instances.shallow()
        if self._metadata is None:
            self._metadata = {}
        return self._metadata

    # copy/paste
    @metadata.setter
    def metadata(self, d):
        self._metadata = {} if d is None else d

    @property
    def instances(self):
        if rwa.islazy(self._instances):
            self._instances = self._instances.shallow()
            if rwa.islazy(self._comments):
                self._comments = self._comments.deep()
            if rwa.islazy(self._metadata):
                self._metadata = self._metadata.deep()
        return InstancesView(self)

    def __str__(self):
        return base.format_analyses(self, node=rwa.lazytype)

    def terminate(self):
        """
        Close the opened file if any and delete all the handles.
        """
        def _terminate(obj, ok=False):
            if rwa.islazy(obj):
                if ok:
                    obj.store.handle = None
                else:
                    obj.store.close()
                    ok = True
            elif isinstance(obj, abc.Analyses):
                obj = obj._instances
                if isinstance(obj, dict): # implicit: not rwa.islazy(obj)
                    for k in obj:
                        ok |= _terminate(obj[k], ok)
                else:
                    ok |= _terminate(obj, ok)
            return ok
        _terminate(self)



def label_paths(analyses, filter, lazy=False):
    """
    Find label paths for analyses matching a criterion.

    Arguments:

        analyses (tramway.core.analyses.base.Analyses):
            hierarchy of analyses, with `instances` possibly containing
            other :class:`~tramway.core.analyses.base.Analyses` instances.

        filter (type or callable):
            criterion over analysis data.

        lazy (bool):
            if applying `filter` function to a :class:`rwa.lazy.LazyPeek`,
            whether to pass the lazy or the evaluated form.

    Returns:

        list of tuples:
            list of label paths to matching analyses.

    """
    if isinstance(filter, type):
        _type = filter
        filter = lambda node: issubclass(rwa.lazytype(node), _type)
    elif callable(filter) and not lazy:
        _filter = filter
        filter = lambda node: _filter(rwa.lazyvalue(node, deep=True))
    return base.label_paths(analyses, filter)


def find_artefacts(analyses, filters, labels=None, quantifiers=None, lazy=False, return_subtree=False):
    """
    Find related artefacts.

    Filters are applied to find data elements (artefacts) along a single path specified by `labels`.

    Arguments:

        analyses (tramway.core.analyses.base.Analyses): hierarchy of analyses.

        filters (type or callable or tuple or list): list of criteria, a criterion being
            a boolean function or a type.

        labels (list): label path.

        quantifiers (str or tuple or list): list of quantifers, a quantifier for now being
            either '*first*', '*last*' or '*all*'; a quantifier should be defined for each
            filter; default is '*last*' (admits value ``None``).

        lazy (bool):
            if applying a filter function to a :class:`rwa.lazy.LazyPeek`,
            whether to pass the lazy or the evaluated form.

        return_subtree (bool): return as extra output argument the analysis subtree corresponding
            to the deepest matching artefact.

    Returns:

        tuple: matching data elements/artefacts, and optionally analysis subtree.

    Examples:

        .. code-block:: python

            cells, maps = find_artefacts(analyses, (CellStats, Maps))

            maps, maps_subtree = find_artefacts(analyses, Maps, return_subtree=True)

    """
    if not isinstance(filters, (tuple, list)):
        filters = (filters,)
    fullnode = lazy or all( isinstance(f, (type, tuple, list)) for f in filters )
    if fullnode:
        # force closure at definition time (otherwise `t` and `f` are overwritten)
        def typefilter(t):
            return lambda a: issubclass(rwa.lazytype(a._data), t)
        def directfilter(f):
            return lambda a: f(rwa.lazyvalue(a, deep=True))
        _filters = []
        for _filter in filters:
            if isinstance(_filter, (type, tuple, list)):
                _type = _filter
                _filter = typefilter(_type)
            elif callable(_filter) and not lazy:
                _filter = directfilter(_filter)
            _filters.append(_filter)
        filters = _filters
    artefacts = base.find_artefacts(analyses, filters, labels, quantifiers, fullnode,
        return_subtree)
    if lazy:
        return artefacts
    else:
        if return_subtree:
            artefacts = list(artefacts)
            subtree = artefacts.pop()
        artefacts = [ rwa.lazyvalue(a, deep=True) for a in artefacts ]
        if return_subtree:
            artefacts.append(subtree)
        return tuple(artefacts)


__all__ = [
    'InstancesView',
    'Analyses',
    'find_artefacts',
    'label_paths',
    ]





# -*- coding: utf-8 -*-

# Copyright © 2020, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from tensorflow.python.keras import saving
from tensorflow import keras
import h5py

class Attr(str):
    __slots__ = ()
    def decode(self, encoding):
        return self
    def __iter__(self):
        if self.startswith('[') and self.endswith(']'):
            for c in self[1:-1].split():
                if c.startswith("b'") and c.endswith("'"):
                    #print(c[2:-1])
                    yield Attr(c[2:-1])
        else:
            for c in str.__iter__(self):
                yield Attr(c)

class AttrsProxy(object):
    __slots__ = ('_attrs',)
    def __init__(self, attrs):
        self._attrs = attrs
    def __contains__(self, attr):
        return attr in self._attrs
    def __getitem__(self, attr):
        assert isinstance(attr, str)
        return Attr(self._attrs[attr])
    def __getattr__(self, attr):
        assert attr not in ('_attrs','__contains__','__getitem__')
        return getattr(self._attrs, attr)

class Proxy(object):
    __slots__ = ('_f',)
    def __init__(self, f):
        self._f = f
    @property
    def attrs(self):
        return AttrsProxy(self._f.attrs)
    def __getitem__(self, attr):
        return self._f[attr]
    def __getattr__(self, attr):
        assert attr not in ('_f','attrs')
        sub = getattr(self._f, attr)
        try:
            sub.attrs
        except AttributeError:
            return sub
        else:
            return Proxy(sub)


__fix_tf_1_14_0_h5py_3_0_0__ = True

class Model(keras.Model):
    def load_weights(self, filepath, by_name=False):
        if __fix_tf_1_14_0_h5py_3_0_0__:
            # originally implemented in tensorflow.python.keras.engine.network
            with h5py.File(filepath, 'r') as f:
                f = Proxy(f)
                if by_name:
                    saving.load_weights_from_hdf5_group_by_name(f, self.layers)
                else:
                    saving.load_weights_from_hdf5_group(f, self.layers)
        else:
            keras.Model.load_weights(self, filepath, by_name)




# -*- coding: utf-8 -*-

# Copyright © 2017, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


import sys
from warnings import warn


def _ro_property_msg(property_name, related_attribute):
    if related_attributed:
        return '`{}` is a read-only property that reflects `{}`''s state'.format(property_name, related_attribute)
    else:
        return '`{}` is a read-only property'.format(property_name)

class PermissionError(AttributeError):
    def __init__(self, property_name, related_attribute):
        RuntimeError.__init__(self, property_name, related_attribute)

    def __str__(self):
        return _ro_property_msg(*self.args)

def ro_property_assert(obj, supplied_value, related_attribute=None, property_name=None, depth=0):
    if property_name is None:
        property_name = sys._getframe(depth + 1).f_code.co_name
    if supplied_value == getattr(obj, property_name):
        warn(_ro_property_msg(property_name, related_attribute), RuntimeWarning)
    else:
        raise PermissionError(property_name, related_attribute)



class Lazy(object):
    """Lazy store.

    Lazily computes and stores attributes through properties, so that the stored attributes can be
    (explicitly) deleted anytime to save memory.

    The :attr:`__lazy__` static attribute is a list of the properties that implement such a
    mechanism.

    Per default each lazy property ``name`` manages a private ``_name`` attribute.
    This naming convention can be overwritten by heriting `Lazy` and overloading
    :meth:`__tolazy__` and :meth:`__fromlazy__` methods.

    An unset lazy attribute/property always has value ``None``.

    A getter will typically look like this:

    .. code-block:: python

        @property
        def name(self):
            if self._name is None:
                self._name = # add some logics
            return self.__lazyreturn__(self._name)

    A fully functional setter will typically look like this:

    .. code-block:: python

        @name.setter
        def name(self, value):
            self.__lazysetter__(value)

    A read-only lazy property will usually look like this:

    .. code-block:: python

        @name.setter
        def name(self, value):
            self.__lazyassert__(value)

    `__lazyassert__` can unset ``_name`` (set it to ``None``) but any other value is treated as
    illegal. `__lazyassert__` compares ``value`` with ``self.name`` and raises a warning if the
    values equal to each other, or throws an exception otherwise.

    """
    __slots__ = ('_lazy',)

    __lazy__  = ()

    def __init__(self):
        self._lazy = {name: True for name in self.__lazy__}
        for name in self.__lazy__:
            setattr(self, self.__fromlazy__(name), None)

    def __returnlazy__(self, name, value):
        return value

    def __lazyreturn__(self, value, depth=0):
        #caller = sys._getframe(depth + 1).f_code.co_name
        #return self.__returnlazy__(caller, value)
        return value

    def __tolazy__(self, name):
        """Returns the property name that corresponds to an attribute name."""
        return name[1:]

    def __fromlazy__(self, name):
        """Returns the attribute name that corresponds to a property name."""
        return '_{}'.format(name)

    def __setlazy__(self, name, value):
        """Sets property `name` to `value`."""
        self._lazy[name] = value is None
        setattr(self, self.__fromlazy__(name), value)

    def __lazysetter__(self, value, depth=0):
        """Sets the property which name is the name of the caller."""
        caller = sys._getframe(depth + 1).f_code.co_name
        self.__setlazy__(caller, value)

    def __assertlazy__(self, name, value, related_attribute=None):
        if value is None: # None has a special meaning for lazy attributes/properties
            self.__setlazy__(name, value)
        else:
            ro_property_assert(self, value, related_attribute, name)

    def __lazyassert__(self, value, related_attribute=None, name=None, depth=0):
        if value is None: # None has a special meaning for lazy attributes/properties
            if name is None:
                self.__lazysetter__(value, depth + 1)
            else:
                self.__setlazy__(name, value)
        else:
            ro_property_assert(self, value, related_attribute, name, depth + 1)

    def unload(self, visited=None):
        """
        Recursively clear the lazy attributes.

        Beware: only direct Lazy object attributes are unloaded,
            not Lazy objects stored in non-lazy attributes!

        *Deprecated*
        """
        warn('`unload` has been removed and will raise an exception in future versions', DeprecationWarning)
        if visited is None:
            visited = set()
        elif id(self) in visited:
            # already unloaded
            return
        visited.add(id(self))
        try:
            names = self.__slots__
        except:
            names = self.__dict__
        standard_attrs = []
        # set lazy attributes to None (unset them so that memory is freed)
        for name in names:
            if self._lazy.get(self.__tolazy__(name), False):
                try:
                    setattr(self, name, None)
                except AttributeError:
                    pass
            else:
                standard_attrs.append(name)
        # search for Lazy object attributes so that they can be unloaded
        for name in standard_attrs: # standard or overwritten lazy
            try:
                attr = getattr(self, name)
            except AttributeError:
                pass
            if isinstance(attr, Lazy):
                attr.unload(visited)



def lightcopy(x):
    """
    Return a copy and call `unload` if available.

    Arguments:

        x (any): object to be copied and unloaded.

    Returns:

        any: copy of `x`.

    *Deprecated*
    """
    warn('`lightcopy` has been removed and will raise an exception in future versions', DeprecationWarning)
    return x


__all__ = [
    'PermissionError',
    'ro_property_assert',
    'Lazy',
    'lightcopy',
    ]




# -*- coding:utf-8 -*-

# Copyright © 2017-2019, Institut Pasteur
#    Contributor: Maxime Duval

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

"""
This module regroups functions generating random distributions.
"""

import numpy as np


def distribution_exp(nb_point=100, lambda_=0.1):
    s = np.random.uniform(0, 1, nb_point)
    return -1 / lambda_ * np.log(1 - s)


def distrib_random_angle(nb_point=1000, dim=1):
    if dim == 1:
        return np.random.randint(2, size=nb_point) * 2 - 1
    elif dim == 2:
        return np.random.uniform(0, 2*np.pi, nb_point)
    elif dim == 3:
        theta = np.random.uniform(0, np.pi, nb_point)
        phi = np.random.uniform(0, 2*np.pi, nb_point)
        return theta, phi
    else:
        raise ValueError(f'dim = {dim} not supported')


def distrib_long_tail_LOMAX_time(alpha=1., lambda_=2., nb_point=100):
    y = np.random.uniform(0, 1, nb_point)
    tau = lambda_*(-1 + np.power(1-y, -1/alpha))
    return tau


def distribution_long_tail_LOMAX_timecut(alpha=1., lambda_=2., tau_max=100,
                                         nb_point=100):

    loc = 1 + tau_max/lambda_
    beta = 1 / (1 - np.power(loc, -alpha))
    y = np.random.uniform(0, 1, nb_point)
    tau = lambda_*(-1 + np.power(beta - y, -1/alpha))
    tau[tau <= 0] = 0
    return tau


def distribution_t_student(df=1, nb_point=1000):
    return np.random.standard_t(df, size=nb_point)


def distribution_uniform_low_max(low_value=0, max_value=1, nb_point=1000):
    s = np.random.uniform(low_value, max_value, 1000)
    return s


def distribution_uniform_low_max_sobol(low_value=0, max_value=1,
                                       nb_point=1000, dim=2):

    s = sobol_seq.i4_sobol_generate(dim, nb_point)
    return low_value + s * (max_value - low_value)


def distrib_long_tail_cst_before_cut(tau_scale=1, alpha=0.25, nb_point=100):
    """
    distribution normalisee (qaund elle est definie cest a dire alpha>1) de
    points avec valeur constante jusqu'à un cut (tau_scale) et ensuite une
    distribution qui chute en t^{-alpha}.
    Distrbution qui permet de ne pas trop définir.
    """
    y = np.random.uniform(0, 1, nb_point)
    y[y <= alpha/(1.+alpha)] = 0.
    y[y > alpha/(1.+alpha)] = 1.
    V_max = 1./(1+alpha)
    # first part
    tau_1 = np.random.uniform(0, tau_scale, nb_point)
    # second part
    y_loc = np.random.uniform(0, V_max, nb_point)
    tau_2 = tau_scale*np.power(1 - (1+alpha)*(V_max - y_loc), -1/alpha)

    tau = (1-y)*tau_1 + y*tau_2
    return tau, tau_1, tau_2


def distrib_long_tail_with_cut_cst_other_cut(tau_scale=0.5, tau_cut=10,
                                             alpha=0.25, nb_point=100):
    """
    distribution normalisee (et definie partout) avec du cut value
    la premiere pour la partie constante de la distribution
    la seconde pour le cut de la partie en  t^{-alpha}
    distrbution qui permet de faire des cas realistes avec des lois d'échelles
    mais un cut evitant les emmerdes et les non normalizations
    tau_scale < tau_cut
    """
    constante1 = tau_scale - np.power(tau_scale, 1. + alpha)/alpha * (
        np.power(tau_cut, -alpha*1.) - np.power(tau_scale, -alpha*1.))
    constante2 = 1/constante1
    constante3 = constante2 * tau_scale
    V_max = -constante2*tau_scale/alpha * \
        (np.power(tau_cut/tau_scale, -alpha*1.)-1)
    y = np.random.uniform(0, 1, nb_point)
    y[y <= constante3] = 0.
    y[y > constante3] = 1.

    # first part
    tau_1 = np.random.uniform(0, tau_scale, nb_point)

    # second part
    y_loc = np.random.uniform(0, V_max, nb_point)
    tau_2 = tau_scale * \
        np.power(1 - alpha/(constante2*tau_scale)*(V_max-y_loc), -1/alpha)

    tau = (1-y)*tau_1 + y*tau_2
    return tau, tau_1, tau_2


def distribution_alpha_levy_stable(alpha=1.5, beta=1, c=1, mu=0,
                                   nb_point=1000):
    """
    distribution alpha-levy stable
    alpha expoenent of the tail part
    beta [-1: 1] skewness parameter
    c [0.. infty[ scale parameter
    mu ]-infty .. infty[
    """

    U = np.random.uniform(-np.pi/2., np.pi/2., nb_point)
    W = distribution_exp(nb_point, 1)
    ksi = -beta * np.tan(np.pi*alpha/2.)

    if alpha != 1:
        eta = 1./alpha*np.arctan(-ksi)

        term1 = np.power(1+ksi*ksi, 1/(2. * alpha))
        term2 = np.sin(alpha*(U + eta))/np.power(np.cos(U), 1./alpha)
        term3 = np.power(np.cos(U - alpha * (U + eta))/W, (1. - alpha)/alpha)

        tau = term1 * term2 * term3
        tau = c*tau + mu

    else:
        eta = np.pi/2.
        term1 = 1./eta
        term2 = (np.pi/2. + beta * U)*np.tan(U)
        term3 = beta*np.log((np.pi/2*W*np.cos(U))/(np.pi/2 + beta*U))

        tau = term1*(term2 - term3)
        tau = c*tau + 2/np.pi*beta*c*np.log(c) + mu

    return tau


def exp_noise(d_max=0.1, nb_point=1000):
    s = np.random.uniform(0, 1, nb_point)
    return -1 / d_max * np.log(1 - s)


def exponential_correlated_noise(dt=0.025, lambda_=1, nb_point=100):
    t = np.arange(nb_point) * dt
    delta_t = np.abs(t[np.newaxis, :] - t[:, np.newaxis])
    covariance = np.exp(-lambda_ * delta_t)
    y = np.random.randn(nb_point)
    L = np.linalg.cholesky(covariance)
    return np.dot(L, y)


def fractional_correlated_noise(dt=0.025, H=0.55, nb_point=100, matrix=False):
    HH = 2 * H
    t = np.arange(nb_point) * dt
    d = np.abs(t[:, np.newaxis] - t[np.newaxis, :])
    covariance = (np.abs(d-dt)**HH + (d+dt)**HH - 2*d**HH) * 0.5
    y = np.random.randn(nb_point)
    L = np.linalg.cholesky(covariance)
    return np.dot(L, y)


def correlation_fractional(sigma=1, H=0.25, t=0, s=1):
    HH = 2 * H
    constante = sigma * sigma / 2
    term1 = np.power(np.absolute(t-s-1), HH)
    term2 = np.power(np.absolute(t-s), HH)
    term3 = np.power(np.absolute(t-s+1), HH)
    return constante * (term1 - 2 * term2 + term3)


def correlation_fractional_t_s(dt=0.025, sigma=1, H=0.25, t_s=0):
    # t_s is an integer
    HH = 2 * H
    constante = sigma * sigma / 2 * np.power(dt, HH)
    term1 = np.power(np.absolute(t_s-1), HH)
    term2 = np.power(np.absolute(t_s), HH)
    term3 = np.power(np.absolute(t_s+1), HH)
    return constante * (term1 - 2*term2 + term3)


def correlation_exponential(sigma=1, lambda_=1, t=0, s=1):
    return sigma_ * np.exp(-lambda_ * np.absolute(t-s))


def fractional_correlated_noise_wood_chan(dt=0.025, sigma=0.5,
                                          H=0.25, nb_point=100):
    nu = int(np.floor(np.log(nb_point) / np.log(2)) + 1)
    m = int(2 * np.power(2, nu))
    m_2 = int(np.floor(m/2))

    e = np.linspace(0, m_2*1, m_2, endpoint=False)
    g = np.linspace(0, (m_2+1)*1, m_2+1, endpoint=False)
    g = g[::-1]
    g = g[0:m_2]
    c2 = np.concatenate((e, g), axis=0)
    c = correlation_fractional_t_s(dt, sigma, H, c2)

    eigen = np.fft.fft(c)

    U = np.random.randn(m_2)
    V = np.random.randn(m_2)

    W = 1j*np.zeros((m,))
    W[0] = U[0]
    W[m_2] = V[0]

    W[1:m_2] = 1 / np.sqrt(2) * (U[1:m_2] + 1j * V[1:m_2])
    W2 = 1 / np.sqrt(2.)*(U[1:m_2] - 1j*V[1:m_2])
    W3 = W2[::-1]
    W[m_2+1:m] = W3

    W2 = 1 / np.sqrt(m) * np.sqrt(eigen) * W
    X = np.fft.fft(W2)
    return X.real


def noisify_gaussian_static(data_in, sigma=0.01):
    return data_in + np.random.randn(data_in.shape) * sigma


def generate_distribution(distribution, nb_point, **kwargs):
    """
    This function is used to generate distributions.
    """
    if distribution == 'exp':
        return distribution_exp(nb_point, 1 / kwargs['d_l'])
    elif distribution == 'uni':
        return np.random.uniform(0, 2 * kwargs['d_l'], nb_point)
    elif distribution == 'gauss':
        if kwargs['dim'] == 1:
            return np.random.normal(0, kwargs['d_l'], nb_point)
        else:
            cst = np.sqrt(np.pi / 2)
            # cst = 1
            return np.abs(np.random.normal(0, kwargs['d_l'] * cst, nb_point))
    elif distribution == 'lomax':
        alpha = kwargs['alpha']
        lambda_ = kwargs['d_l'] * (alpha - 1) if alpha > 1 else 1
        return distrib_long_tail_LOMAX_time(alpha, lambda_, nb_point)
    elif distribution == 'constant_and_tail':
        alpha = kwargs['alpha']
        d_scale = kwargs['d_l']
        tau_scale = d_scale * (1 + alpha) / (alpha * (1/2 + 1 / (alpha - 1)))
        return distrib_long_tail_cst_before_cut(tau_scale, alpha, nb_point)[0]
    elif distribution == "alpha_stable":
        args = (kwargs['alpha'], 0, kwargs['c_alpha_levy'], 0, nb_point)
        return np.abs(distribution_alpha_levy_stable(*args))
    elif distribution == 'lomax_cut':
        prms = (kwargs['alpha'], kwargs['d_l'], kwargs['d_max'], nb_point)
        return distribution_long_tail_LOMAX_timecut(*prms)
    elif distribution == 'constant_tail_cut':
        prms = (kwargs['alpha'], kwargs['d_l'], kwargs['d_max'], nb_point)
        return distrib_long_tail_with_cut_cst_other_cut(*prms)[0]
    elif distribution == 'cst':
        return np.ones(nb_point) * kwargs['d_l']


def generate_times(distribution, nb_point, T_max, dt, alpha=1.5,
                   d_tau=0.02, d_tau_max=1, dim=2, c_scale_alpha_stable_t=1):
    if distribution == "cst":
        t = np.linspace(0, nb_point*dt, nb_point, endpoint=False)
    else:
        t = generate_distribution(distribution, nb_point - 1, alpha=alpha,
                                  d_l=d_tau, d_max=d_tau_max, dim=dim,
                                  c_alpha_levy=c_scale_alpha_stable_t)
        while np.sum(t) < T_max:
            t_add = generate_distribution(distribution, nb_point - 1,
                                          alpha=alpha,
                                          d_l=d_tau, d_max=d_tau_max, dim=dim,
                                          c_alpha_levy=c_scale_alpha_stable_t)
            t = np.concatenate((t, t_add))
        t = np.insert(t.cumsum(), 0, 0)
    return t


# For diffusing diffusivity

def give_diffusion_chubynsky_slater(nb_point=1000, dt=0.025,
                                    sigma_bruit=0.1, g_bruit=0.1,
                                    D_init=0.05):
    # equaton dD/dt = -g + sigma*xi(t)
    # reflecting boundary conditions at D = 0
    # equilibirum diffusion D* = sigma^2/(2*g)
    # equilibirum distribution 1/D*exp(-D/D*)
    noise_ = sigma_bruit*np.sqrt(dt)*np.random.randn(nb_point, 1)
    D = np.zeros((nb_point,))
    D[0] = D_init
    drift = -g_bruit*dt
    for i in range(1, nb_point):
        D[i] = D[i-1] + drift + noise_[i]
        if D[i] < 0:
            D[i] *= -1
    return D


def give_diffusion_langevin_square(n_dimension=1, nb_point=1000, dt=0.025,
                                   tau_bruit=10, sigma_bruit=0.1, D_init=0.05):
    # 1D most likely in our case could be extended to ndimention : to our
    # application a bit pointless
    # Y is auxiliary variable
    noise_ = sigma_bruit*np.sqrt(dt)*np.random.randn(nb_point, n_dimension)
    Y = np.zeros((nb_point, n_dimension))
    Y[0, :] = 1/np.sqrt(n_dimension)*np.sqrt(D_init)*np.ones((1, n_dimension))
    const = (1 - dt/tau_bruit)
    for i in range(1, nb_point):
        Y[i, :] = const * Y[i-1, :] + noise_[i-1, :]
        Y[i, Y[i, :] < 0] = -Y[i,  Y[i, :] < 0]
    D = np.power(np.linalg.norm(Y, axis=1), 2)
    return D


def give_diffusion_CIR(nb_point=1000, dt=0.025, tau_bruit=10, sigma_bruit=0.1,
                       D_init=0.05):
    # modele de Cox–Ingersoll–Ross utilise par grebenkov dans
    # J. Phys. A: Math. Theor. 51 145602
    noise_ = sigma_bruit*np.sqrt(dt)*np.random.randn(nb_point, 1)
    D = np.zeros((nb_point,))
    D[0] = D_init
    const = (1. - dt/tau_bruit)
    drift = D_init/tau_bruit*dt
    for i in range(1, nb_point):
        D[i] = const*D[i-1] + drift + np.sqrt(2*D[i-1])*noise_[i-1]
        if D[i] < 0:
            D[i] *= -1
    return D




import numpy as np
from skimage import draw
from math import *

def line_aa(i0, j0, i1, j1, thickness=1):
    if thickness:
        thickness = max(1., thickness)
        # grid
        i_lb = ceil(min(i0, i1) - thickness)
        j_lb = ceil(min(j0, j1) - thickness)
        i_ub = floor(max(i0, i1) + thickness)
        j_ub = floor(max(j0, j1) + thickness)
        i = np.arange(i_lb, i_ub+1)
        j = np.arange(j_lb, j_ub+1)
        i_, j_ = np.meshgrid(i, j, indexing='ij')
        i_, j_ = i_.ravel(), j_.ravel()
        ij = np.stack((i_, j_), axis=1)
        # distance to segment
        di, dj = i1 - i0, j1 - j0
        n = sqrt(di * di + dj * dj)
        n_ = np.r_[dj / n, -di / n]
        ij0 = np.array([[i0, j0]])
        d = np.dot(ij - ij0, n_)
        d2 = d * d
        d_max = 1. * thickness
        ok = d2 < d_max * d_max
        # selected indices and corresponding alpha
        i = i_[ok]
        j = j_[ok]
        d2 = d2[ok]
        d = np.sqrt(d2)
        d_cutoff = .5 * thickness
        k = 1 / (1 + np.exp(4 * (d - d_cutoff) / d_cutoff))
        # discard the corners
        w_ = np.r_[di / n, dj / n]
        w = np.dot(ij[ok] - ij0, w_)
        w0 = w<0
        w1 = n<w
        d0 = -w[w0]
        d0 = np.sqrt(d0 * d0 + d2[w0])
        d1 = w[w1] - n
        d1 = np.sqrt(d1 * d1 + d2[w1])
        k[w0] = np.minimum(k[w0],
                1 / (1 + np.exp(4 * (d0 - d_cutoff) / d_cutoff)))
        k[w1] = np.minimum(k[w1],
                1 / (1 + np.exp(4 * (d1 - d_cutoff) / d_cutoff)))
        i, j = i.astype(dtype=int), j.astype(dtype=int)
    else:
        i, j, k = draw.line_aa(i0, j0, i1, j1)
    return i, j, k
    
def multiline_aa(path_i, path_j, thickness=0):
    """
    If thickness is null, pixel indices and line thickness should be integers.
    Otherwise, floating-point values are allowed.

    Arguments:

        path_i (array-like): pixel column indices

        path_j (array-like): pixel row indices

        thickness (int): thickness in pixels

    Returns:

        (ndarray, ndarray, ndarray): column indices, row indices, pixel values

    """
    max_k = {}
    for i0, j0, i1, j1 in \
            zip(path_i[:-1], path_j[:-1], path_i[1:], path_j[1:]):
        for i, j, k in zip(*line_aa(i0, j0, i1, j1, thickness)):
            try:
                k_prev = max_k[(i,j)]
            except KeyError:
                pass
            else:
                k = max(k, k_prev)
            max_k[(i,j)] = k
    i, j, k = zip(*[ ij+(k,) for ij, k in max_k.items() ])
    return np.array(i), np.array(j), np.array(k)

__all__ = ['line_aa', 'polyline_aa']



# -*- coding: utf-8 -*-

# Copyright © 2020, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from ..abc import Initializer
from . import merger as cell_mergers

class TessellationPostProcessingInitializer(Initializer):
    """
    Initializer class for attribute :attr:`~tramway.analyzer.RWAnalyzer.tesseller`
    :attr:`~tramway.analyzer.tesseller.proxy.TessellerProxy.post`.

    For now, a single use case is available:

    .. code-block:: python

        a = RWAnalyzer()
        # initialize attribute `tesseller`
        a.tesseller = tessellers.KMeans
        # .. so that attribute `post` is exposed
        a.tesseller.post = cell_mergers.ByTranslocationCount
        # .. and now `tesseller.post` is initialized
        a.tesseller.post.count_threshold = 20

    Note that :mod:`~tramway.analyzer.cell_mergers` is exported by the
    :mod:`tramway.analyzer` module, just like :mod:`~tramway.analyzer.tesseller.tessellers`.

    See also :class:`~tramway.analyzer.tesseller.post.merger.ByTranslocationCount`.
    """
    __slots__ = ()
    def from_callable(self, cls):
        """
        Argument:

            cls (callable):
                no-argument callable object (usually a class) that returns a
                :class:`~tramway.analyzer.tesseller.TessellationPostProcessing` object.

        """
        self.specialize( cls )




"""
Functions allowing to calculate the D posterior and MAP(D) as (will be) described in the Ito-Stratonovich article.
Provides 3 functions:
- get_D_posterior
- get_MAP_D
- get_D_confidence_interval

If unsure, use `get_D_confidence_interval`, which provides MAP(D) and a confidence interval for D for the given confidence level.
If need the D posterior, use `get_D_posterior`.
"""


import logging
import warnings

import numpy as np
from numpy import exp as exp
from numpy import log as log
from scipy.optimize import brentq
from scipy.special import gammainc, gammaincc, gammaln

from .convenience_functions import n_pi_func
from .convenience_functions import p as pow

# prior assumption for diffusivity inference that does not include D'.
# Should be described in the article Appendix


def get_D_posterior(n, zeta_t, V, V_pi, dt, sigma2, dim, _MAP_D=False):
    """Return the diffusivity posterior as a function.
    If _MAP_D flag is supplied, returns instead a single MAP(D) value

    Input:
    V   -   (biased) variance of jumps in the current bin
    V_pi    -   (biased) variance of jumps in all other bins excluding the current one
    dt  -   time step
    sigma2  -   localization error (in the units of variance)
    dim -   dimensionality of the problem
    _MAP_D  -   if True, returns MAP_D instead of the posterior

    Output:
    posterior    -   function object accepting D as the only argument
    """

    n_pi = n_pi_func(dim)
    p = pow(n, dim)
    v = 1.0 + n_pi / n * V_pi / V
    eta = np.sqrt(n_pi / (n + n_pi))
    G3 = v + eta**2 * _norm2(zeta_t - _zeta_mu(dim))

    if _MAP_D:
        return n * V * G3 / 4 / dt / (p + 1)

    # Conversions
    zeta_t = np.array(zeta_t)
    rel_loc_error = n * V / (4 * sigma2) if sigma2 > 0 else np.inf

    def posterior(D):
        """Remember gammainc is normalized to gamma"""
        min_D = sigma2 / dt
        if D <= min_D:
            ln_res = -np.inf
        else:
            ln_res = (p * log(n * V * G3 / (4 * dt * D))
                      - n * V * G3 / (4 * D * dt)
                      - log(D)
                      - log(gammainc(p, rel_loc_error * G3))
                      - gammaln(p))
        return np.exp(ln_res)
    return posterior


def get_MAP_D(n, zeta_t, V, V_pi, dt, sigma2, dim):
    return get_D_posterior(n, zeta_t, V, V_pi, dt, sigma2, dim, _MAP_D=True)


def get_D_confidence_interval(alpha, n, zeta_t, V, V_pi, dt, sigma2, dim):
    """Returns MAP_D and a confidence interval corresponding to confidence level alpha, i.e. the values of D giving the values [(1-alpha)/2, 1-(1-alpha)/2] for the D posterior integrals

    Input:
    alpha   -   confidence level,
    n       -   number of jumps in bin,
    zeta_t  -   signal-to-noise ratio for the total force: <dr>/sqrt(V),
    V       -   (biased) variance of jumps in the current bin,
    V_pi    -   (biased) variance of jumps in all other bins excluding the current one,
    dt      -   time step
    sigma2  -   localization error (in the units of variance),
    dim     -   dimensionality of the problem

    Output:
    MAP_D   -   MAP value of the diffusivity,
    CI      -   a confidence interval for the diffusivity, 2x1 np.array

    """
    n_pi = n_pi_func(dim)
    p = pow(n, dim)
    v = 1.0 + n_pi / n * V_pi / V
    eta = np.sqrt(n_pi / (n + n_pi))
    zeta_t = np.array(zeta_t)
    G3 = v + eta**2 * _norm2(zeta_t - _zeta_mu(dim))
    min_D = sigma2 / dt    # units of variance /s
    MAP_D = get_MAP_D(n, zeta_t, V, V_pi, dt, sigma2, dim)
    y_L = n * V * G3 / 4 / sigma2

    def posterior_integral(D):
        y = n * V * G3 / 4 / dt / D
        if D <= min_D:
            intg = 0
        else:
            intg = 1 - gammainc(p, y) / gammainc(p, y_L)
        return intg

    # %% Find root
    # Maximal D up to which to look for a root, units of variance /s
    # Calculated as asymptotic root times 2
    # TODO may result in error if low alpha is given as input
    q = 1 - (1 - alpha) / 2
    max_D_search = (-1 / p) * (log(p) + log(1 - q) + log(gammainc(p, y_L)) - gammaln(p))
    max_D_search = n * V * G3 / 4 / dt / exp(max_D_search)
    max_D_search *= 2
    max_D_search = np.max([max_D_search, 1e3])

    CI = np.ones(2) * np.nan
    for i, q in enumerate([(1 - alpha) / 2, 1 - (1 - alpha) / 2]):

        def solve_me(D):
            return posterior_integral(D) - q
        CI[i] = brentq(solve_me, min_D, max_D_search)

    return MAP_D, CI


def _zeta_mu(dim):
    """The center of the prior for the total force.
    For diffusivity inference should not depend on the diffusivity gradient.
    """
    return np.zeros(dim)


def _norm2(x):
    return np.sum(x**2)



# -*- coding: utf-8 -*-

# Copyright © 2018, Alexander Serov

# from log_C import log_C as log_C_func
import numpy as np
from scipy import optimize

from .calculate_marginalized_integral import calculate_marginalized_integral
from .convenience_functions import p


def find_marginalized_zeta_t_roots(zeta_sp_par, n, n_pi, B, u, dim, zeta_t_perp, sigma2):
    """
    Find marginalized roots zeta_t.
    I have proven that the min B for the marginalized inference is achieved at zeta_t = zeta_sp/2 (for the parallel component and under the condition that the orthogonal component is 0).
    sigma2 --- localization error. Same units as variance. Set to 0 if localization error can be ignored;
    """

    # Constants
    max_init_search_attempts = 50
    increase_factor = 2.0

    # # Check input
    # if not isinstance(zeta_sp, list):
    # 	raise TypeError("'zeta_sp' must be a 1D list.")

    eta = np.sqrt(n_pi / (n + n_pi))
    # p = dim * (n + n_pi + 1.0) / 2.0 - 2.0
    pow = p(n, dim)

    # # Define v function
    # def v(s):
    #     return(1.0 + n_pi / n * u + (dim - 1.0) * s * zeta_t_perp ** 2)
    v = 1.0 + n_pi / n * u

    rel_loc_error = n * V / (4 * sigma2) if sigma2 > 0 else np.inf

    # Function to optimize
    def solve_me(zeta_t_par):
        zeta_sp = np.asarray([zeta_sp_par, 0])
        zeta_t = np.asarray([zeta_t_par, zeta_t_perp])

        E = eta ** 2.0
        # print([zeta_t_cur], [zeta_sp], p, v(E), E)
        # calculate_marginalized_integral takes [2,1] vectors as input that include the parallel and othogonal components
        upstairs = calculate_marginalized_integral(zeta_t, zeta_sp, pow, v, E, rel_loc_error)

        E = 1.0
        downstairs = calculate_marginalized_integral(zeta_t, zeta_sp, pow, v, E, rel_loc_error)
        return upstairs * eta ** dim - B * downstairs

    # Guess sign change intervals centered around zeta_t = zeta_sp / 2
    zeta_t_lims = np.array([0.0, 0.5, 1.0]) * zeta_sp_par
    zeta_t_width = zeta_t_lims[-1] - zeta_t_lims[0]
    min_value = solve_me(zeta_t_lims[1])

    # If the minimized function is negative for the middle point we won't be able to find roots
    # print(min_value)
    # print([solve_me(zeta_t_lims[0]), solve_me(zeta_t_lims[-1])])
    if min_value > 0:
        # print("Warning: No marginalized zeta_t roots exist for zeta_sp = %.2f. Min Bayes factor value (%.3g) is always greater than the requested value (%.3g) for any zeta_t." % (zeta_sp, min_value + K, K))
        return [np.nan, np.nan]

    attempt = 0
    bl_interval_found = False
    while(attempt < max_init_search_attempts):
        # Check if the end points have a different sign from the middle
        end_values = [solve_me(zeta_t_lims[0]), solve_me(zeta_t_lims[-1])]

        if end_values[0] * min_value < 0 and end_values[1] * min_value < 0:
            # The interval found
            bl_interval_found = True
            break
        else:
            # Increase interval
            zeta_t_width = zeta_t_width * increase_factor
            zeta_t_lims[0] = zeta_t_lims[1] - 0.5 * zeta_t_width
            zeta_t_lims[-1] = zeta_t_lims[1] + 0.5 * zeta_t_width
        attempt += 1

    # Check if the interval was found
    if not bl_interval_found:
        raise AssertionError(
            "Failed to converge to initial search interval for the marginalized roots")

    # print (end_values, min_value)
    # print(zeta_t_lims)
    # print(attempt)

    # If the initial intervals found, find the two roots
    zeta_t_roots = []

    # Find the smaller root
    for i in [0, 1]:
        zeta_t_roots.append(optimize.brentq(
            solve_me, zeta_t_lims[i], zeta_t_lims[i + 1]))

    # Sort roots
    zeta_t_roots.sort()
    # print(zeta_t_roots)

    return(zeta_t_roots)



# -*- coding: utf-8 -*-

# Copyright © 2017,2019, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


import numpy as np
import pandas as pd
from .namedcolumns import *


class Scaler(object):
    """:class:`Scaler` scales data points, point differences (vectors) or distances.

    It initializes itself with the first provided sample (in :meth:`scale_point`), and then applies
    the same transformation to the next samples.

    A default ``Scaler()`` instance does not scale. However, initialization still takes place
    so that :meth:`scaled` properly works.

    It manages a constraint in the calculation of the scaling parameters, forcing a common factors
    over a subset of dimensions. Attribute :attr:`euclidean` controls the selection of this subset.
    Distances are scaled and unscaled only in this subspace, if it is defined.

    Beware that when possible data are scaled in place, but `scaledonly` optional argument, when
    available, never operates in place.

    Attributes:
        init (bool):
            ``True`` as long as `Scaler` has not been initialized.
        center (array or pandas.Series):
            Vector that is substracted to each row of the data matrix to be scaled.
        factor (array or pandas.Series):
            Vector by which each row of the data matrix to be scaled is divided.
            Applies after `center`.
        columns (list or pandas.Index):
            Sequence of column names along which scaling applies. This applies only to
            structured data. `columns` is determined even if `Scaler` is set to do nothing,
            so that :meth:`scaled` can still apply.
            `columns` can be manually set after the first call to :meth:`scale_point` if data
            are not structured (do not have named columns).
        function (callable):
            A function that takes a data matrix as input and returns `center` and `factor`.
            `function` is called once during the first call to :meth:`scale_point`.
        euclidean (list):
            Sequence of names or indices of the columns to be scaled by a common factor.
    """
    __slots__ = ['init', 'center', 'factor', 'columns', 'function', 'euclidean']

    def __init__(self, scale=None, euclidean=None):
        """
        Arguments:
            scale (callable):
                A function that takes a data matrix as input and returns `center` and
                `factor`. `scale` becomes the :attr:`function` attribute.
            euclidean (list):
                Sequence of names or indices of the columns to be scaled by a common
                factor.
        """
        self.init   = True
        self.center = None
        self.factor = None
        self.columns = []
        self.function = scale
        if euclidean and not \
            (isinstance(euclidean, list) and euclidean[1:]):
            raise TypeError('`euclidean` should be a multi-element list')
        self.euclidean = euclidean

    @property
    def ready(self):
        """Returns `True` if scaler is initialized."""
        return not self.init

    def scaled(self, points, asarray=False):
        """Discard columns that are not recognized by the initialized scaler.

        Applies to points and vectors, not distances, surface areas or volumes."""
        if len(self.columns):
            if isstructured(points):
                cols = columns(points)
                coltype = type(cols[0])
                if type(self.columns[0]) is not coltype:
                    if coltype is bytes:
                        coerce = lambda s: s.encode('utf-8')
                    else:
                        coerce = lambda s: s.decode('utf-8')
                    self.columns = [ coerce(c) for c in self.columns ]
                    self.euclidean = [ coerce(c) for c in self.euclidean ]
                points = points[self.columns]
            else:
                try:
                    points = np.asarray(points)[:, self.columns]
                except IndexError:
                    cols = list(self.columns)
                    raise TypeError("the input data array does not feature the following columns: {}".format(cols))
        elif isstructured(points):
            raise ValueError("input data are structured whereas scaler's internal data are not")
        else:
            scaler_data = self.center
            if scaler_data is None:
                scaler_data = self.factor
            if scaler_data is None:
                if self.function:
                    raise RuntimeError('scaler has not been initialized')
            else:
                if isinstance(points, (tuple, list)):
                    points = np.asarray(points)
                if scaler_data.shape[1] != points.shape[1]:
                    raise ValueError('number of columns does not match')
        if asarray:
            points = np.asarray(points)
        return points

    def scale_point(self, points, inplace=True, scaledonly=False, asarray=False):
        """
        Scale data.

        When this method is called for the first time, the `Scaler` instance initializes itself
        for further call of any of its methods.

        Arguments:
            points (array-like):
                Data matrix to be scaled. When :meth:`scale_point` is called for the
                first time, `points` can be structured or not, without the unnecessary
                columns, if any.
                At further calls of any (un-)scaling method, the input data should be in
                the same format with at least the same column names, and may feature
                extra columns.
            inplace (bool):
                Per default, scaling is performed in-place. With ``inplace=False``,
                `points` are first copied.
            scaledonly (bool):
                If ``True``, undeclared columns are stripped away out of the returned
                data.
            asarray (bool):
                If ``True``, the returned data is formatted as a :class:`numpy.array`.

        Returns:
            array-like: With default optional input arguments, the returned variable will be
                a pointer to `points`, not otherwise.
        """
        if isinstance(points, (tuple, list)):
            points = np.asarray(points)
        if self.init:
            # define named columns
            if self.columns:
                raise AttributeError('remove data columns at initialization instead of defining `columns`')
            try:
                self.columns = list(columns(points)) # structured arrays do not support tuple indexing
            except:
                pass
            # backup predefined values
            if self.center is None:
                predefined_centers = []
            elif isinstance(self.center, list):
                predefined_centers = self.center
            if self.factor is None:
                predefined_factors = []
            elif isinstance(self.factor, list):
                predefined_factors = self.factor
            if self.function:
                # calculate centers and factors
                self.center, self.factor = self.function(points)
                # equalize factor for euclidean variables
                if self.euclidean:
                    if isinstance(points, (pd.Series, pd.DataFrame)):
                        xyz = points[self.euclidean].values
                    elif points.dtype.names:
                        xyz = np.asarray(points[list(self.euclidean)])
                    else:
                        xyz = np.asarray(points)[:,self.euclidean]
                    _, self.factor[self.euclidean] = self.function(xyz.flatten())
            # overwrite the coordinates that were actually predefined
            if predefined_centers:
                if self.center is None:
                    self.center = __get_row(points, 0.0)
                for col, val in predefined_centers:
                    self.center[col] = val
            if predefined_factors:
                if self.factor is None:
                    self.factor = __get_row(points, 1.0)
                for col, val in predefined_factors:
                    self.factor[col] = val
            self.init = False
        if not (self.center is None and self.factor is None):
            points = _format(points, inplace)
            if isinstance(points, np.ndarray):
                if self.center is not None:
                    points -= np.asarray(self.center)
                if self.factor is not None:
                    points /= np.asarray(self.factor)
            else:
                if self.center is not None:
                    points -= self.center
                if self.factor is not None:
                    points /= self.factor
        if scaledonly:
            points = self.scaled(points, asarray)
        elif asarray:
            points = np.asarray(points)
        return points

    def unscale_point(self, points, inplace=True):
        """
        Scale data back to original domain.

        The calling `Scaler` instance must have been initialized.

        Arguments:
            points (array-like):
                Scaled data matrix to be unscaled.
            inplace (bool):
                Per default, scaling is performed in-place. With ``inplace=False``,
                `points` are first copied.

        Returns:
            array-like: unscaled data matrix.
        """
        if self.init:
            raise AttributeError('scaler has not been initialized')
        if not (self.center is None and self.factor is None):
            points = _format(points, inplace)
            if isinstance(points, np.ndarray):
                if self.factor is not None:
                    points *= np.asarray(self.factor)
                if self.center is not None:
                    points += np.asarray(self.center)
            else:
                if self.factor is not None:
                    points *= self.factor
                if self.center is not None:
                    points += self.center
        return points


    def scale_vector(self, vect, inplace=True, scaledonly=False, asarray=False):
        """
        Scale vectors.

        The calling `Scaler` instance must have been initialized.

        Arguments:
            vect (array-like):
                Data matrix to be scaled.
            inplace (bool):
                Per default, scaling is performed in-place. With ``inplace=False``,
                `vect` is first copied.
            scaledonly (bool):
                If ``True``, undeclared columns are stripped away out of the returned
                data.
            asarray (bool):
                If ``True``, the returned data is formatted as a :class:`numpy.array`.

        Returns:
            array-like: scaled data matrix.
        """
        if self.init:
            raise AttributeError('scaler has not been initialized')
        if self.factor is not None:
            vect = _format(vect, inplace)
            vect /= self.factor
        if scaledonly:
            vect = self.scaled(vect, asarray)
        elif asarray:
            vect = np.asarray(vect)
        return vect

    def unscale_vector(self, vect, inplace=True):
        """
        Scale vectors back to original range.

        The calling `Scaler` instance must have been initialized.

        Arguments:
            vect (array-like):
                Scaled data matrix to be unscaled.
            inplace (bool):
                Per default, scaling is performed in-place. With ``inplace=False``,
                `points` are first copied.

        Returns:
            array-like: unscaled data matrix.
        """
        if self.init:
            raise AttributeError('scaler has not been initialized')
        if self.factor is not None:
            vect = _format(vect, inplace)
            vect *= self.factor
        return vect

    def scale_size(self, size, dim=None, inplace=True, _unscale=False):
        """
        Scale/unscale lengths, surface areas, volumes and other scalar sizes.

        The calling `Scaler` instance must have been initialized.

        Arguments:
            size (array-like):
                Values to be scaled, per element.
            dim (int):
                Number of characteristic dimensions, with 0 referring to all the
                euclidean dimensions (e.g. lengths: 1, areas: 2, volumes: 0).
            inplace (bool):
                Per default, scaling is performed in-place. With ``inplace=False``,
                `size` is first copied.
            _unscale (bool):
                If ``True``, unscales instead.

        Returns:
            array-like: scaled values.
        """
        if self.init:
            raise AttributeError('scaler has not been initialized')
        if self.factor is not None:
            _dim = len(self.euclidean)
            if not dim:
                dim = _dim
            if _dim < min(1, dim):
                raise ValueError('not enough euclidean dimensions')
            try:
                factor = self.factor[self.euclidean[0]]
            except KeyError:
                # on loading Py2-generated Series or DataFrame from an rwa file,
                # PyTables may convert Py2 str in index/columns into Py3 str;
                # as a consequence `columns` and `euclidean` should also be converted
                self.columns = [ name.decode('utf-8') for name in self.columns ]
                self.euclidean = [ name.decode('utf-8') for name in self.euclidean ]
                factor = self.factor[self.euclidean[0]]
            if self.euclidean[1:] and not np.all(self.factor[self.euclidean[1:]] == factor):
                raise ValueError('the scaling factors for the euclidean variables are not all equal')
            size = _format(size, inplace)
            if 1 < dim:
                factor **= dim
            if _unscale:
                size *= factor
            else:
                size /= factor
        return size

    def scale_distance(self, dist, inplace=True):
        return self.scale_size(dist, 1, inplace)

    def unscale_distance(self, dist, inplace=True):
        return self.scale_size(dist, 1, inplace, True)

    def scale_length(self, dist, inplace=True):
        return self.scale_size(dist, 1, inplace)

    def unscale_length(self, dist, inplace=True):
        return self.scale_size(dist, 1, inplace, True)

    def scale_surface_area(self, area, inplace=True):
        return self.scale_size(area, 2, inplace)

    def unscale_surface_area(self, area, inplace=True):
        return self.scale_size(area, 2, inplace, True)

    def scale_volume(self, vol, inplace=True):
        return self.scale_size(vol, 0, inplace)

    def unscale_volume(self, vol, inplace=True):
        return self.scale_size(vol, 0, inplace, True)


def _whiten(x):
    '''Scaling function for :class:`Scaler`. Performs ``(x - mean(x)) / std(x)``. Consider using
    :func:`whiten` instead.'''
    scaling_center = x.mean(axis=0)
    scaling_factor = x.std(axis=0, ddof=0)
    return (scaling_center, scaling_factor)

def whiten(): # should be a function so that each new instance is a distinct one
    """Returns a :class:`Scaler` that scales data ``x`` following: ``(x - mean(x)) / std(x)``."""
    return Scaler(_whiten)


def _unitrange(x):
    '''Scaling function for :class:`Scaler`. Performs ``(x - min(x)) / (max(x) - min(x))``.
    Consider using :func:`unitrange` instead.'''
    scaling_center = x.min(axis=0)
    scaling_factor = x.max(axis=0) - scaling_center
    return (scaling_center, scaling_factor)

def unitrange():
    """Returns a :class:`Scaler` that scales data ``x`` following:
    ``(x - min(x)) / (max(x) - min(x))``."""
    return Scaler(_unitrange)


def __get_row(points, fill=None):
    if isinstance(points, (pd.Series, pd.DataFrame)):
        row = points.iloc[0]
    else:
        row = points[0]
    if fill is not None:
        if isinstance(points, (tuple, list)):
            row = np.asarray(row)
        row.fill(fill)
    return row


def _format(points, inplace):
    if inplace:
        if isinstance(points, tuple):
            raise TypeError('cannot modify tuple inplace')
        elif isinstance(points, list):
            raise NotImplementedError('cannot modify list inplace')
    else:
        if isinstance(points, (tuple, list)):
            points = np.asarray(points)
        else:
            points = points.copy()
    return points


__all__ = [
    'Scaler',
    'whiten',
    'unitrange',
    ]




# -*- coding: utf-8 -*-

# Copyright © 2018, Alexander Serov
"""
The file contains posterior calculation functions.
"""

import numpy as np
from numpy import exp, log, pi
from scipy import integrate
from scipy.special import gamma, gammainc, gammaln

from .calculate_marginalized_integral import (calculate_any_lambda_integral,
                                              calculate_integral_ratio,
                                              calculate_marginalized_integral)
from .convenience_functions import n_pi_func
from .convenience_functions import p as pow


def calculate_one_2D_posterior(zeta_t, zeta_sp, zeta_a, n, V, V_pi, loc_error, dim=2, lamb='marg'):
    """
    Calculate the posterior for zeta_a for a given method for one bin in 2D

    Input:
    lambda --- float in [0, 1] or 'marg' --- for different conventions
    """

    # Check if None is present
    if any(var is None for var in [zeta_t, zeta_sp, n, V, V_pi, loc_error]):
        logging.info('None values encountered in cell. Skipping cell.')
        return [np.nan] * 3

    if lamb == 'marg':
        lamb = 'int'

    if loc_error > 0:
        rel_loc_error = n * V / (4 * loc_error)
    else:
        rel_loc_error = np.inf

    # Parameter combinations
    n_pi = n_pi_func(dim)
    p = pow(n, dim)
    p_upstairs = p + dim / 2
    u = V_pi / V
    v = 1.0 + n_pi / n * u
    eta = np.sqrt(n_pi / (n + n_pi))
    factor_za = eta**2 / (1 - eta**2)

    prefactor = (np.pi * (1 - eta**2))**(-dim / 2)

    upstairs = calculate_marginalized_integral(zeta_t=zeta_t, zeta_sp=zeta_sp,
                                               p=p_upstairs, v=v, E=1.0, rel_loc_error=rel_loc_error, zeta_a=zeta_a, factor_za=factor_za, lamb=lamb)
    downstairs = calculate_marginalized_integral(zeta_t=zeta_t, zeta_sp=zeta_sp,
                                                 p=p, v=v, E=eta**2, rel_loc_error=rel_loc_error, zeta_a=[0, 0], factor_za=0, lamb=lamb)
    gamma_ratio = gamma(p_upstairs) / gamma(p)
    res = prefactor * upstairs / downstairs * gamma_ratio

    return res


def calculate_one_1D_posterior_in_2D(zeta_a, zeta_t, zeta_sp, n, V, V_pi, loc_error, lamb='marg', axis='x'):
    """
    Calculate the posterior for a projection of zeta_a for a given inference convention for one bin in 2D

    Input:
    lambda: float in [0, 1] or 'marg'; for different conventions
    axis: ['x', 'y']; for which projection to calculate the posterior
    zeta_a: float; `a` projection on the selected axis
    """
    dim = 2
    # Check if None is present
    if any(var is None for var in [zeta_t, zeta_sp, n, V, V_pi, loc_error]):
        logging.info('None values encountered in cell. Skipping cell.')
        return [np.nan] * 3

    zeta_t, zeta_sp, zeta_a = map(np.asarray, [zeta_t, zeta_sp, zeta_a])
    i_post, i_marg = (0, 1) if axis == 'x' else (1, 0)
    rel_loc_error = n * V / (4 * loc_error) if loc_error > 0 else np.inf

    # Parameter combinations
    n_pi = n_pi_func(dim)
    p = pow(n, dim)
    p_upstairs = p + 1 / 2
    v = 1.0 + n_pi * V_pi / n / V
    log_prefactor = (log(n + n_pi) - log(n) - log(pi)) / 2

    def upstairs(l):
        arg = (v
               + n_pi / n * zeta_a**2
               + (zeta_t[i_post] - zeta_a - l * zeta_sp[i_post])**2
               + n_pi / (n + n_pi) * (zeta_t[i_marg] - l * zeta_sp[i_marg])**2)
        return arg

    def downstairs(l):
        arg = (v
               + n_pi / (n + n_pi) * np.sum((zeta_t - l * zeta_sp)**2))
        return arg

    # Identify break points
    with np.errstate(invalid='ignore', divide='ignore'):
        bps = set([(zeta_t[i_post] - zeta_a) / zeta_sp[i_post],
                   zeta_t[i_marg] / zeta_sp[i_marg]])
        bps |= set(zeta_t / zeta_sp)
    bps = [bp for bp in bps if bp >= 0 and bp <= 1]

    # Final calculations
    log_ratio = calculate_integral_ratio(arg_func_up=upstairs, arg_func_down=downstairs,
                                         pow_up=p_upstairs, pow_down=p, v=v,
                                         rel_loc_error=rel_loc_error, break_points=bps,
                                         lamb=lamb)
    return exp(log_prefactor + log_ratio)


def calculate_one_1D_prior_in_2D(zeta_a, V_pi, sigma2):
    """
    Calculate the prior for a projection of zeta_a for any convention for one bin in 2D

    Input:
    zeta_a: float; `a` projection on the selected axis
    """
    dim = 2

    zeta_a = np.asarray(zeta_a)
    n_pi = n_pi_func(dim)
    rel_loc_error = n_pi * V_pi / (4 * sigma2) if sigma2 > 0 else np.inf

    # Parameter combinations
    p = n_pi - 2
    p_upstairs = p + 1 / 2
    log_prefactor = - log(pi) / 2

    def upstairs(l):
        return 1 + zeta_a**2

    def downstairs(l):
        return 1

    log_ratio = calculate_integral_ratio(arg_func_up=upstairs, arg_func_down=downstairs,
                                         pow_up=p_upstairs, pow_down=p, v=1,
                                         rel_loc_error=rel_loc_error, break_points=[], lamb=0)
    r = log_prefactor + log_ratio
    return exp(r)


def get_lambda_MAP(zeta_t, zeta_sp):
    """
    Calculate the maximum a posteriori value of lambda for given zetas in 1D and 2D
    """
    lamb = np.dot(zeta_t, zeta_sp) / np.linalg.norm(zeta_sp)**2

    # Limit to the region where the prior is defined
    if lamb > 1:
        lamb = 1
    if lamb < 0:
        lamb = 0

    return lamb



# -*- coding: utf-8 -*-

# Copyright © 2019-2021, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


try:
    import cloudpickle
except ImportError:
    pass
import multiprocessing
try:
    import queue
except ImportError:
    import Queue as queue
import time
import numpy as np
from warnings import warn
try:
    from . import abc
except SyntaxError: # Py2
    import abc_py2 as abc
from traceback import format_exc

#import logging # DEBUG
#module_logger = logging.getLogger(__name__)
#module_logger.setLevel(logging.DEBUG)
#_console = logging.StreamHandler()
#_console.setFormatter(logging.Formatter('%(message)s\n'))
#module_logger.addHandler(_console)


class StarQueue(object):
    """
    A star queue is a multidirectional queue such that each message must be consumed
    by every processes but the sender.
    This is useful to send asynchronous updates between processes in a distributed setting.

    A :class:`StarQueue` is instanciated first, and then :class:`StarConn` objects are delt
    to children processes.
    The :class:`StarConn` objects are the actual queues.
    """
    __slots__ = 'deck',
    def __init__(self, n, variant=multiprocessing.Queue, **kwargs):
        self.deck = []
        queues = [ variant(**kwargs) for _ in range(n) ]
        others = []
        for _ in range(n):
            _queue = queues.pop()
            self.deck.append((_queue, queues+others))
            others.append(_queue)
    def deal(self):
        try:
            return StarConn(*self.deck.pop())
        except IndexError:
            raise RuntimeError('too many consumers')

class StarConn(object):
    __slots__ = 'input', 'output'
    def __init__(self, input_queue=None, output_queues=None):
        self.input = input_queue
        self.output = output_queues
    @property
    def joinable(self):
        return isinstance(self.input, multiprocessing.JoinableQueue)
    @property
    def variant(self):
        return type(self.input)
    def qsize(self):
        return sum( q.qsize() for q in self.output )
    def empty(self):
        return all( q.empty() for q in self.output )
    def full(self):
        return any( q.full()  for q in self.output)
    def put(self, obj, block=True, timeout=None):
        if timeout and 0 < timeout:
            t0 = time.time()
        _timeout = timeout
        for _queue in self.output:
            _queue.put(obj, block, _timeout)
            if timeout and 0 < timeout:
                _timeout = timeout - (time.time() - t0)
    def put_nowait(self, obj):
        for _queue in self.output:
            _queue.put_nowait(obj)
    def get(self, block=True, timeout=None):
        return self.input.get(block, timeout)
    def get_nowait(self):
        return self.input.get_nowait()
    def close(self):
        if self.output is None:
            raise RuntimeError('queue not yet `close` should not be called by the parent process')
        for _queue in self.output:
            _queue.close()
    def join_thread(self):
        if self.output is None:
            raise RuntimeError('`join_thread` should not be called by the parent process')
        for _queue in self.output:
            _queue.join_thread()
    def task_done(self):
        if self.input is None:
            raise RuntimeError('no task was received through the queue')
        try:
            self.input.task_done()
        except (SystemExit, KeyboardInterrupt):
            raise
        except:
            if self.joinable:
                raise
            else:
                raise RuntimeError('queue was not set as joinable at init')

class WorkerNearDeathException(Exception):
    __slots__ = '_id', '_name', '_type', '_msg'
    def __init__(self, _id, name, exc_type, traceback):
        self._id = _id
        self._name = name
        self._type = exc_type
        self._msg = traceback.split('\n',1)[1][:-1] # remove first and last lines
    def __str__(self):
        return 'Process {} died with error (most recent call last):\n{}'.format(self._name, self._msg)

class Worker(multiprocessing.Process):
    """ Worker that runs job steps.

    The :meth:`target` method may be implemented following the pattern below:

    .. code:: python

        class MyWorker(Worker):
            def target(self, *args, **kwargs):
                while True:
                    k, task = self.get_task()
                    status = dict()
                    try:
                        # modify `task` and `status`
                    finally:
                        self.push_update(task, status)

    The optional positional and keyword arguments come from the `args` and `kwargs` arguments
    to :meth:`Scheduler.__init__`, plus the extra keyword arguments to the latter constructor.

    """
    def __init__(self, _id, workspace, task_queue, return_queue, update_queue,
            name=None, args=(), kwargs={}, daemon=None, **_kwargs):
        # `daemon` is not supported in Py2; pass `daemon` only if defined
        if daemon is None:
            __kwargs = {}
        else:
            __kwargs = dict(daemon=daemon)
        multiprocessing.Process.__init__(self, name=name, **__kwargs)
        self._id = _id
        self.workspace = workspace
        self.tasks = task_queue
        self.update = update_queue
        self.feedback = return_queue
        self.args = args
        kwargs = dict(kwargs) # copy
        kwargs.update(_kwargs)
        self.kwargs = kwargs
    def get_task(self):
        """ Listen to the scheduler and get a job step to be run.

        The job step is loaded with the worker-local copy of the synchronized workspace.

        Returns:

            int, tramway.core.parallel.JobStep: step/iteration number, job step object.

        """
        #module_logger.debug('get_task: waiting...') # DEBUG
        k, task = self.tasks.get()
        #module_logger.debug('get_task: received {}'.format(k)) # DEBUG
        task.set_workspace(self.workspace)
        self.pull_updates()
        return k, task
    def push_update(self, update, status=None):
        """ Send a completed job step back to the scheduler and to the other workers.

        Arguments:

            update (tramway.core.parallel.JobStep): completed job step.

            status (any): extra information that :meth:`Scheduler.stop` will receive.

        """
        if isinstance(update, abc.VehicleJobStep):
            update.push_updates(self.workspace.pop_extension_updates())
        update.unset_workspace() # free memory space
        if self.update is not None:
            self.update.put(update)
        #module_logger.debug('push_update: sending back') # DEBUG
        self.feedback.put((update, status))
        #module_logger.debug('push_update: sent') # DEBUG
    def pull_updates(self):
        if self.update is None:
            return
        while True:
            try:
                update = self.update.get_nowait()
            except queue.Empty:
                break
            else:
                self.workspace.update(update) # `Workspace.update` reloads the workspace into the update
    def run(self):
        try:
            self.target(*self.args, **self.kwargs)
        except (SystemExit, KeyboardInterrupt):
            raise
        except Exception as e:
            self.feedback.put((None,
                WorkerNearDeathException(self._id, self.name, type(e), format_exc())))

class NormalTermination(Exception):
    pass

def _pseudo_worker(worker):
    class PseudoWorker(worker):
        def __init__(self, scheduler, args=(), kwargs={}, _id=0, name=None):
            worker.__init__(self, _id, scheduler.workspace,
                None, None, None, name=name, args=args, kwargs=kwargs)
            self._scheduler = scheduler
        def get_task(self):
            return self._scheduler.next_task()
        def push_update(self, update, status=None):
            i = update.resource_id
            self._scheduler.task[i] = update
            if self._scheduler.pseudo_stop(status):
                raise NormalTermination
        def run(self):
            self.target(*self.args, **self.kwargs)
    return PseudoWorker


class Scheduler(object):
    """ Scheduler that distributes job steps over a shared workspace.

    Workers are spawned and assigned job steps.
    Each worker maintains a copy of the common workspace that is synchronized
    by :meth:`Worker.push_update` calls.

    The :meth:`stop` method should be overloaded so that the distributed computation
    may complete on termination criteria.
    """
    def __init__(self, workspace, tasks, worker_count=None, iter_max=None,
            name=None, args=(), kwargs={}, daemon=None, max_runtime=None,
            task_timeout=None, **_kwargs):
        """
        Arguments:

            workspace (Workspace): workspace to be replicated

            ...

            max_runtime (float): timeout in seconds;
                runtime is absolute, not cumulated over all workers;
                the active workers are interrupted on reaching the timeout;
                *new in 0.5b2*.

            task_timeout (float): timeout in seconds;
                this timeout applies to each task and the corresponding
                worker only is interrupted;
                the other tasks keep on;
                *new in 0.5b5*.

        """
        self.workspace = workspace
        self.task = tasks
        self.active = dict()
        self.paused = dict()
        self.dead_workers = dict()
        self.k_eff = 0
        self.k_max = iter_max
        self.global_timeout = max_runtime
        self.task_timeout = task_timeout
        if worker_count is None:
            worker_count = multiprocessing.cpu_count() - 1
        elif worker_count < 0:
            worker_count = multiprocessing.cpu_count() + worker_count
        kwargs = dict(kwargs) # copy
        kwargs.update(_kwargs)
        if worker_count:
            self.task_queue = multiprocessing.Queue()
            self.return_queue = multiprocessing.Queue()
            if worker_count == 1:
                self.workers = { 0: self.worker(0, self.workspace,
                            self.task_queue, self.return_queue, None,
                            name=name, args=args, kwargs=kwargs, daemon=daemon) }
            else:
                def _name(w):
                    return '{}-{:d}'.format(name, w) if w else None
                update_queue = StarQueue(worker_count)
                self.workers = { i: self.worker(i, self.workspace,
                            self.task_queue, self.return_queue, update_queue.deal(),
                            name=_name(i), args=args, kwargs=kwargs, daemon=daemon)
                        for i in range(worker_count) }
        else:
            self.workers = self.pseudo_worker(self, args=args, kwargs=kwargs)
    def init_resource_lock(self):
        if 1 < self.worker_count:
            self.resource_lock = np.zeros(len(self.workspace), dtype=bool)
    @property
    def worker(self):
        return Worker
    @property
    def pseudo_worker(self):
        return _pseudo_worker(self.worker)
    @property
    def worker_count(self):
        return len(self.workers) if isinstance(self.workers, dict) else 0
    @property
    def timeout(self):
        if self.task_timeout is None:
            return self.global_timeout
        elif self.global_timeout is None:
            return self.task_timeout
        else:
            return min(self.global_timeout, self.task_timeout)
    def draw(self, k):
        return k
    def next_task(self):
        """ no-mp mode only """
        task = None
        while not self.iter_max_reached():
            i = self.draw(self.k_eff)
            self.k_eff += 1
            if i is not None:
                task = self.task[i]
                break
        return self.k_eff-1, task
    def locked(self, step):
        return 1 < self.worker_count and \
                (step.resource_id in self.active or \
                np.any(self.resource_lock[step.resources]))
    def lock(self, step):
        if 1 < self.worker_count:
            self.resource_lock[step.resources] = True
    def unlock(self, step):
        if 1 < self.worker_count:
            self.resource_lock[step.resources] = False
    @property
    def available_slots(self):
        """
        `int`: Number of available workers.
        """
        return self.worker_count - len(self.active)
    def send_task(self, k, step):
        """
        Send a job step to be assigned to a worker as soon as possible.

        Arguments:

            k (int): step/iteration number.

            step (tramway.core.parallel.JobStep): job step.

        """
        self.active[step.resource_id] = k
        self.lock(step)
        step.unset_workspace() # free memory
        #module_logger.debug('send_task: sending {}...'.format(k)) # DEBUG
        self.task_queue.put((k, step))
        #module_logger.debug('send_task: sent') # DEBUG
        self.k_eff += 1
    def get_processed_step(self):
        """
        Retrieve a processed job step and check whether stopping criteria are met.

        Calls the :meth:`stop` method.

        Returns :const:`False` if a stopping criterion has been met, :const:`True` otherwise.
        """
        while True:
            #module_logger.debug('get_processed_step: waiting...') # DEBUG
            # note: `get` does not raise TimeoutError
            step, status = self.return_queue.get(timeout=self.timeout)
            #module_logger.debug('get_processed_step: received {}'.format(self.active[step.resource_id])) # DEBUG
            if step is None:
                if isinstance(status, WorkerNearDeathException):
                    self.dead_workers[status._id] = self.workers.pop(status._id, None)
                    try:
                        self.logger.critical(str(status))
                    except AttributeError:
                        print('error: {}'.format(status))
                    if not self.workers:
                        return False
                else:
                    return self.stop(None, None, status)
            else:
                break
        #step.set_workspace(self.workspace) # reload workspace
        self.workspace.update(step) # `update` reloads the workspace into `step`
        assert step.get_workspace() is not None
        i = step.resource_id
        self.task[i] = step
        k = self.active.pop(i)
        try:
            return not self.stop(k, i, status)
        finally:
            self.unlock(step)
    def iter_max_reached(self):
        return self.k_max and self.k_max <= self.k_eff
    def workers_alive(self):
        self.workers = { i: w for i, w in self.workers.items() if w.is_alive() }
        return bool(self.workers)
    def fill_slots(self, k, postponed):
        """
        Send as many job steps as there are available workers.

        Arguments:

            k (int) : step/iteration number.

            postponed (dict): postponed job steps.

        Returns:

            int: new step/iteration number.

        """
        while 0 < self.available_slots:
            i = self.draw(k)
            if i is None:
                k += 1
            elif i in self.active or i in postponed:
                break
            else:
                task = self.task[i]
                if self.locked(task):
                    postponed[i] = k
                else:
                    self.send_task(k, task)
                k += 1 # increment anyway in the case i is None
                if self.iter_max_reached(): # can happen only after `send_task`
                    #assert False
                    break
        return k
    def run(self):
        """
        Start the workers, send and get job steps back and check for stop criteria.

        Returns :const:`True` on normal completion, :const:`False` on interruption
        (:class:`SystemExit`, :class:`KeyboardInterrupt`).
        """
        if isinstance(self.workers, Worker):
            # no multi-processing
            context = None
            if self.timeout:
                try:
                    import stopit
                except ImportError:
                    warn('limited timeout support; install package stopit to get full support', ImportWarning)
                else:
                    context = stopit.ThreadingTimeout(self.timeout)
            if context is None:
                class DummyContextManager(object):
                    def __enter__(self):
                        return self
                    def __exit__(self, *args):
                        pass
                context = DummyContextManager()
            #
            ret = True
            try:
                with context:
                    self.workers.run()
            except NormalTermination:
                pass
            except (SystemExit, KeyboardInterrupt):
                ret = False
            return ret

        for w in self.workers.values():
            w.start()
        self.init_resource_lock()
        if self.global_timeout:
            self.start_time = time.time()
        k = 0
        postponed = dict()
        try:
            k = self.fill_slots(k, postponed)
            while not self.iter_max_reached():
                if not self.workers_alive():
                    break
                if not self.get_processed_step():
                    break
                if self.global_timeout and self.global_timeout <= (time.time() - self.start_time):
                    #raise multiprocessing.TimeoutError
                    break
                if postponed:
                    for i in list(postponed):
                        task = self.task[i]
                        if not self.locked(task):
                            self.send_task(postponed.pop(i), task)
                            if self.available_slots == 0:
                                break
                            else:
                                assert 0 < self.available_slots
                if not postponed:
                    k = self.fill_slots(k, postponed)
            ret = True
        except (SystemExit, KeyboardInterrupt, multiprocessing.TimeoutError):
            ret = False
        for w in self.workers.values():
            try:
                w.terminate()
            except:
                pass
        for w in self.dead_workers.values():
            try:
                w.terminate()
            except:
                pass
        return ret
    def stop(self, k, i, status):
        """
        Default implementation returns ``False``.

        Arguments:

            k (int): step/interation number.

            i (int): step id.

            status (dict): status data returned by a job step.

        Returns:

            bool: ``True`` if a stopping criterion has been met, ``False`` otherwise.

        """
        return False
    def pseudo_stop(self, status):
        k = self.k_eff - 1
        return self.stop(k, self.draw(k), status)


class EpochScheduler(Scheduler):
    def __init__(self, workspace, tasks, epoch_length=None, soft_epochs=False, worker_count=None,
            iter_max=None, name=None, args=(), kwargs={}, daemon=None, **_kwargs):
        epoch_length = len(tasks) if epoch_length is None else epoch_length
        if not soft_epochs and not worker_count:
            worker_count = min(epoch_length, multiprocessing.cpu_count() - 1)
        Scheduler.__init__(self, workspace, tasks, worker_count=worker_count, iter_max=iter_max,
            name=name, args=args, kwargs=kwargs, daemon=daemon, **_kwargs)
        self.soft_epochs = soft_epochs
        self._task_epoch = np.arange(epoch_length)

    def draw(self, k):
        i = k % len(self._task_epoch)
        if i == 0:
            # wait for the active buffer to get empty
            if not self.soft_epochs and self.active:
                return None
            # permute the tasks
            self.start_new_epoch(self._task_epoch)
        return self._task_epoch[i]

    def start_new_epoch(self, task_order):
        """ must modify the `task_order` array inplace """
        np.random.shuffle(task_order)


class ProtoWorkspace(object):
    __slots__ = '_extensions',
    def __init__(self, args=()):
        self._extensions = {}
        if args:
            self.identify_extensions(args)
    def update(self, step):
        step.set_workspace(self)
        if isinstance(step, abc.VehicleJobStep):
            self.push_extension_updates(step.pop_updates())
    def resources(self, step):
        return step.resources
    def identify_extensions(self, args):
        for k, a in enumerate(args):
            if isinstance(a, abc.WorkspaceExtension):
                self._extensions[k] = a
    def push_extension_updates(self, updates):
        for k in updates:
            update = updates[k]
            extension = self._extensions[k]
            extension.push_workspace_update(update)
    def pop_extension_updates(self):
        updates = {}
        for k in self._extensions:
            extension = self._extensions[k]
            updates[k] = extension.pop_workspace_update()
        return updates

abc.ExtendedWorkspace.register(ProtoWorkspace)


class Workspace(ProtoWorkspace):
    """ Parameter singleton.

    Attributes:

        data_array (array-like): working copy of the parameter vector.

    """
    __slots__ = 'data_array',
    def __init__(self, data_array, *args):
        ProtoWorkspace.__init__(self, args)
        self.data_array = data_array
    def __len__(self):
        return len(self.data_array)


class JobStep(object):
    """ Job step data.

    A job step object contains all the necessary input data for a job step
    to be performed as well as the output data resulting from the step completion.

    A job step object merely contains a reference to a shared workspace.

    The `resource_id` attribute refers to a series of job steps that operate
    on the same subset of resource items.

    Multiple steps can operate simultaneously in the same workspace in a distributed fashion
    provided that they do not compete for the same resources.

    `resources` is an index array that designates the items of shared data to be accessed.
    This attribute is used by :class:`Scheduler` to lock the required items of data,
    which determines which steps can be run simultaneously.
    """
    __slots__ = '_id', '_workspace'
    def __init__(self, _id, workspace=None):
        self._id = _id
        self._workspace = workspace
    def get_workspace(self):
        return self._workspace
    def set_workspace(self, ws):
        self._workspace = ws
    def unset_workspace(self):
        self._workspace = None
    @property
    def workspace_set(self):
        return self._workspace is not None
    @property
    def resource_id(self):
        return self._id
    @property
    def resources(self):
        return self.get_workspace().resources(self)

abc.JobStep.register(JobStep)


class UpdateVehicle(object):
    """ Not instanciable! Introduced for __slots__-enabled multiple inheritance.

    Example usage, in the case class ``B`` implements (abc.) :class:`VehicleJobStep` and
    class ``A`` can only implement (abc.) :class:`JobStep` and not
    inherit from :class:`VehiculeJobStep`::

        class A:
            __slots__ = 'a',
        abc.JobStep.register(A)
        class B(A, UpdateVehicle):
            __slots__ = ('b', ) + VehicleJobStep.__slots__
            def __init__(self, a, b):
                A.__init__(self, a)
                UpdateVehicle.__init__(self)
                self.b = b
        abc.VehicleJobStep.register(B)

    :class:`VehicleJobStep` brings the slots, :class:`UpdateVehicle` brings the implementation
    (methods) and :class:`abc.VehicleJobStep` the typing required by :class:`Workspace` and
    :class:`Worker` to handle ``B`` as a :class:`VehicleJobStep`.
    """
    __slots__ = ()
    def __init__(self):
        self._updates = {}
    def pop_updates(self):
        try:
            return self._updates
        finally:
            self._updates = {}
    def push_updates(self, updates):
        self._updates = updates

class VehicleJobStep(JobStep, UpdateVehicle):
    __slots__ = '_updates',
    def __init__(self, _id, workspace=None):
        JobStep.__init__(self, _id, workspace)
        UpdateVehicle.__init__(self)

abc.VehicleJobStep.register(VehicleJobStep)


__all__ = [ 'StarConn', 'StarQueue', 'ProtoWorkspace', 'Workspace', 'JobStep', 'UpdateVehicle', 'VehicleJobStep', 'Worker', 'Scheduler', 'EpochScheduler', 'abc' ]

